from dataclasses import dataclass
from typing import Optional

from . import CameraDistortion


@dataclass
class CameraIntrinsics:
    fx: float
    fy: float
    cx: float
    cy: float
    width: int
    height: int

    distortion: Optional[CameraDistortion]
    skew: float = 0
    scale_factor: float = 0



from __future__ import annotations

from functools import reduce
from typing import Iterator, Optional, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from numpy.typing import ArrayLike
from scipy.interpolate import make_interp_spline
from scipy.spatial.transform import Rotation

from .spacial_path import SpacialPath, SpacialInterpolationData

IDENTITY = (0, 0, 0, 0, 0, 0, 1)


class PosePath(SpacialPath):
    """PosePath class representing a list of poses at given timestamps, extending pandas DataFrame."""

    XYZ = ["x", "y", "z"]
    QUAT = ["qx", "qy", "qz", "qw"]
    COLUMNS = XYZ + QUAT

    def __init__(
        self, data: Union[ArrayLike, pd.DataFrame], index: Optional[ArrayLike] = None
    ):
        """Initializes the PosePath object.

        Args:
            data (Union[ArrayLike, pd.DataFrame]): An array or DataFrame of poses with shape (N, 7), where N is the number of poses.
            index (ArrayLike, optional): An array-like object representing the index (the timestamps) for the PosePath DataFrame. Defaults to None.
        """
        super().__init__(data=data, index=index, columns=PosePath.COLUMNS, dtype=float)

    def copy(self) -> PosePath:
        """Creates a copy of the current PosePath object.

        Returns:
            PosePath: A new PosePath object with copied data and index.
        """
        return PosePath(self.values.copy(), index=self.index)

    @classmethod
    def from_rt(
        cls,
        rotation: Rotation,
        translation: ArrayLike,
        index: Optional[ArrayLike] = None,
    ) -> PosePath:
        """Creates a PosePath object from rotations and translations.

        Args:
            rotation (Rotation): A Rotation object representing the rotations.
            translation (ArrayLike): An array-like object of translations with shape (N, 3), where N is the number of translations.
            index (ArrayLike, optional): An array-like object representing the index for the PosePath DataFrame. Defaults to None.

        Returns:
            PosePath: A PosePath object with the given rotation and translation.
        """
        positions = np.asarray(translation).reshape((-1, 3))
        headings = rotation.as_quat().reshape((-1, 4))
        assert len(headings) == len(positions)
        return PosePath(np.hstack([positions, headings]), index=index)

    @classmethod
    def from_csv(cls, file: str) -> PosePath:
        """Creates a PosePath object from a CSV file.

        Args:
            file (str): The path to the CSV file.

        Returns:
            PosePath: A PosePath object with data read from the CSV file.
        """
        return PosePath(pd.read_csv(file, index_col=0))

    @classmethod
    def identity(cls, n: int = 1, index: Optional[ArrayLike] = None) -> PosePath:
        """Create a PosePath object with identity poses.

        Args:
            n (int, optional): The number of identity poses. Defaults to 1.
            index (ArrayLike, optional): An array-like object representing the index for the PosePath DataFrame. Defaults to None.

        Returns:
            PosePath: A PosePath object with identity poses.
        """
        return PosePath(np.tile(IDENTITY, n).reshape((n, 7)), index=index)

    @classmethod
    def from_matrix(
        cls, matrix: ArrayLike, index: Optional[ArrayLike] = None
    ) -> PosePath:
        """Creates a PosePath object from transformation matrices.

        Args:
            matrix (ArrayLike): A 3D array-like object of transformation matrices with shape (N, 4, 4), where N is the number of matrices.
            index (ArrayLike, optional): An array-like object representing the index for the PosePath DataFrame. Defaults to None.

        Returns:
            PosePath: A PosePath object with poses represented by the given transformation matrices.
        """
        matrix = np.asarray(matrix).reshape((-1, 4, 4))
        return PosePath.from_rt(
            Rotation.from_matrix(matrix[:, :3, :3]), matrix[:, :3, 3], index=index
        )

    def as_matrix(self) -> npt.NDArray:
        """Convert the PosePath object to transformation matrices.

        Returns:
            npt.NDArray: A 3D array of transformation matrices with shape (N, 4, 4), where N is the number of poses.
        """
        matrix = np.tile(np.eye(4), (len(self), 1, 1))
        matrix[:, :3, :3] = Rotation.from_quat(self.headings).as_matrix()
        matrix[:, :3, 3] = self.positions
        return matrix

    @classmethod
    def from_euler(cls, seq: str, angles: ArrayLike, degrees: bool = False) -> PosePath:
        """Creates a PosePath object from Euler angles.

        Args:
            seq (str): The Euler sequence of axes, such as 'xyz', 'zyx', etc.
            angles (ArrayLike): An array-like object of Euler angles with shape (N, len(seq)), where N is the number of poses.
            degrees (bool, optional): If True, angles are in degrees. Defaults to False.

        Returns:
            PosePath: A PosePath object with poses represented by the given Euler angles.
        """
        angles = np.asarray(angles).reshape((-1, len(seq)))
        path = PosePath.identity(n=len(angles))
        path.headings = Rotation.from_euler(seq, angles, degrees).as_quat()
        return path

    def as_euler(self, seq: str, degrees: bool = False) -> npt.NDArray:
        """Converts the PosePath object to Euler angles.

        Args:
            seq (str): The Euler sequence of axes, such as 'xyz', 'zyx', etc.
            degrees (bool, optional): If True, angles are in degrees. Defaults to False.

        Returns:
            np.ndarray: An array of Euler angles with shape (N, len(seq)), where N is the number of poses.
        """
        return Rotation.from_quat(self.headings).as_euler(seq, degrees=degrees)

    @classmethod
    def from_positions(cls, positions: ArrayLike) -> PosePath:
        """Creates a PosePath object with given positions.

        Args:
            positions (ArrayLike): An array-like object of positions with shape (N, 3), where N is the number of positions.

        Returns:
            PosePath: A PosePath object with poses represented by the given positions and identity orientations.
        """
        positions = np.asarray(positions).reshape((-1, 3))
        path = PosePath.identity(len(positions))
        path.positions = positions
        return path

    @property
    def positions(self) -> npt.NDArray:
        """Gets the positions of the poses.

        Returns:
            npt.NDArray: An array of positions with shape (N, 3), where N is the number of poses.
        """
        return self.values[:, 0:3]

    @positions.setter
    def positions(self, values: ArrayLike) -> None:
        """Set the positions of the poses.

        Args:
            values (ArrayLike): Anarray-like object of positions with shape (N, 3), where N is the number of positions.
        """
        self.values[:, 0:3] = np.asarray(values).reshape((-1, 3))

    @property
    def headings(self) -> npt.NDArray:
        """Gets the orientations (headings) of the poses in quaternions.

        Returns:
            npt.NDArray: An array of quaternions with shape (N, 4), where N is the number of poses.
        """
        return self.values[:, 3:7]

    @headings.setter
    def headings(self, values: ArrayLike) -> None:
        """Sets the orientations (headings) of the poses in quaternions.

        Args:
            values (ArrayLike): An array-like object of quaternions with shape (N, 4), where N is the number of orientations.
        """
        self.values[:, 3:7] = np.asarray(values).reshape((-1, 4))

    # Operations
    def invert(self) -> PosePath:
        """Creates a new PosePath instance with inverted poses.

        Returns:
            PosePath: A PosePath object with inverted poses.
        """
        inv_rotations = Rotation.from_quat(self.headings).inv()
        inv_positions = -inv_rotations.apply(self.positions)
        return PosePath.from_rt(inv_rotations, inv_positions, index=self.index)

    def __matmul__(self, other: Union[PosePath, ArrayLike]) -> PosePath:
        """Matrix multiplication of the PosePath object with another PosePath object or a transformation matrix.

        Args:
            other (Union['PosePath', ArrayLike]): Another PosePath object or a transformation matrix/array.

        Returns:
            PosePath: A PosePath object with poses resulting from the matrix multiplication.
        """
        if isinstance(other, PosePath):
            resampled = other.interpolate(self.index)
            return PosePath.from_matrix(
                self.as_matrix() @ resampled.as_matrix(), index=self.index
            )

        return PosePath.from_matrix(
            self.as_matrix() @ np.asarray(other), index=self.index
        )

    def __rmatmul__(self, other: Union[PosePath, ArrayLike]) -> PosePath:
        """Right matrix multiplication of the PosePath object with another PosePath object or a transformation matrix.

        Args:
            other (Union['PosePath', ArrayLike]): Another PosePath object or a transformation matrix/array.

        Returns:
            PosePath: A PosePath object with poses resulting from the matrix multiplication.
        """
        if isinstance(other, PosePath):
            resampled = other.interpolate(self.index)
            return PosePath.from_matrix(
                resampled.as_matrix() @ self.as_matrix(), index=self.index
            )

        return PosePath.from_matrix(
            np.asarray(other) @ self.as_matrix(), index=self.index
        )

    @classmethod
    def multiply(cls, paths: Iterator[PosePath]) -> PosePath:
        """Composes multiple PosePath objects.

        Args:
            paths (Iterator['PosePath']): An iterator of PosePath objects.

        Returns:
            PosePath: A PosePath object with poses resulting from the matrix multiplication of the given PosePath objects.
        """
        return reduce(cls.__rmatmul__, paths)

    def make_relative(self) -> PosePath:
        """Creates a new PosePath object with poses relative to the first pose.

        Returns:
            PosePath: A new PosePath object with poses relative to the first pose in the original PosePath object.
        """
        inv_first = self[:1].invert().as_matrix()[0]
        return inv_first @ self

    def __getitem__(self, key):
        result = super().__getitem__(key)
        if isinstance(result, pd.Series):
            return result
        elif isinstance(result, pd.DataFrame) and np.array_equal(
            result.columns, PosePath.COLUMNS
        ):
            return PosePath(result)
        return result

    # InterpolateablePath
    @classmethod
    def from_interpolation_data(
        cls, value: SpacialInterpolationData, index: npt.NDArray
    ) -> PosePath:
        """Creates a PosePath from interpolated values

        Parameters
        ----------
        values : SpacialInterpolationData
            Interpolated position and rotation
        index : npt.NDArray
            Timestamps

        Returns
        -------
        PosePath
            Result of interpolation.
        """
        assert (
            value.spherical is not None and value.linear is not None
        ), f"Unexpected None value in data {value}"
        return PosePath.from_rt(value.spherical, value.linear, index=index)

    def get_interpolation_data(self) -> SpacialInterpolationData:
        """Gets the values of this object for interpolation.

        Returns
        -------
        SpacialInterpolationData
            Position and rotation.
        """
        return SpacialInterpolationData(
            linear=self.positions, spherical=Rotation.from_quat(self.headings)
        )

    def calculate_kinematics(self, time_unit_to_second_conversion: int = 1e9) -> pd.DataFrame:
        """Calculate the velocity and acceleration vectors of the pose path by
            fitting a spline to the positions and computing 1st and 2nd
            derivatives for each dimension. Also converts the timestamps to
            seconds (assumes nanoseconds by default).

        Returns:
            pd.DataFrame: A pandas DataFrame with the xyz components of the
                            velocity and acceleration of the pose path at each
                            timestamp with columns [vx, vy, vz, ax, ay, az].
        """
        timestamps_in_seconds = self.index / time_unit_to_second_conversion

        def calculate_1d_kinematics(axis: int) -> (npt.NDArray, npt.NDArray):
            spline = make_interp_spline(timestamps_in_seconds, self.positions[:, axis])
            velocities = spline(timestamps_in_seconds, nu=1)
            accelerations = spline(timestamps_in_seconds, nu=2)
            return velocities, accelerations

        x_velocities, x_accelerations = calculate_1d_kinematics(0)
        y_velocities, y_accelerations = calculate_1d_kinematics(1)
        z_velocities, z_accelerations = calculate_1d_kinematics(2)

        return pd.DataFrame({
            'vx': x_velocities,
            'vy': y_velocities,
            'vz': z_velocities,
            'ax': x_accelerations,
            'ay': y_accelerations,
            'az': z_accelerations
        }, index=timestamps_in_seconds)



from dataclasses import dataclass
from typing import List, Union, Optional, Literal

from ..common import AnnotationID, AnnotationKind, SensorID
from ..paths import AttributePath

# Define AttributesAnnotation dataclass
@dataclass
class AttributesAnnotation:
    id: AnnotationID
    type: Literal[AnnotationKind.Attributes] = AnnotationKind.Attributes
    parent_id: Optional[AnnotationID] = None
    attributes: Optional[List[AttributePath]] = None




from dataclasses import dataclass
from typing import List, Optional, Dict, Any

from typing_extensions import Literal

from .annotations import Annotation
from .annotations.attributes import AttributePath
from .sensors import Sensor


@dataclass
class Scene:
    sensors: Optional[List[Sensor]] = None
    annotations: Optional[List[Annotation]] = None
    attributes: Optional[List[AttributePath]] = None
    time_offset: Optional[int] = None
    time_unit: Optional[
        Literal["microseconds", "nanoseconds"]
    ] = "microseconds"
    metadata: Optional[Dict[str, Any]] = None

    def add_sensor(self) -> None:
        pass




from dataclasses import dataclass
from typing import List, Literal, Optional

from ..common import AnnotationID, AnnotationKind, SensorID
from ..paths import AttributePath

@dataclass
class ObjectAnnotation:
    id: AnnotationID
    type: Literal[AnnotationKind.Object] = AnnotationKind.Object
    parent_id: Optional[AnnotationID] = None
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None




from enum import Enum
from typing import Optional, cast

import dacite
from scale_json_binary import JSONBinaryEncoder

from ..spec import BS5

encoder = JSONBinaryEncoder()


def _fix_data_shape(scene: BS5.Scene) -> BS5.Scene:
    """
    When reading via dacite, the numpy arrays aren't correctly shaped (since that's not included in the typedef)
    This function fixes all the fields that need to be reshaped
    """
    if scene.sensors:
        for sensor in scene.sensors:
            if sensor.type == "lidar":
                for l_frame in sensor.frames:
                    l_frame.points.positions = l_frame.points.positions.reshape(-1, 3)
                    if l_frame.points.colors is not None:
                        l_frame.points.colors = l_frame.points.colors.reshape(-1, 3)

            elif sensor.type == "radar":
                for r_frame in sensor.frames:
                    r_frame.points.positions = r_frame.points.positions.reshape(-1, 3)
                    if r_frame.points.directions is not None:
                        r_frame.points.directions = r_frame.points.directions.reshape(
                            -1, 3
                        )
                    if r_frame.points.lengths is not None:
                        r_frame.points.lengths = r_frame.points.lengths.reshape(-1, 3)

            elif sensor.type == "points":
                sensor.points.positions = sensor.points.positions.reshape(-1, 3)
    return scene


class BS5Loader:
    def __init__(
        self,
        scene_url: str,
    ):
        self.scene_url = scene_url

    def load(self) -> BS5.Scene:
        raw_data: bytes
        with open(self.scene_url, "rb") as fd:
            raw_data = cast(bytes, fd.read())
        return self.load_from_bytes(raw_data)
        

    @staticmethod
    def load_from_bytes(raw_data: bytes) -> BS5.Scene:
        obj = encoder.loads(raw_data)
        if "version" not in obj or not obj["version"].startswith("5.1") and not obj["version"].startswith("1.0"):
            raise Exception(f"Cannot load scene with version {obj['version']}")

        scene_bs5 = dacite.from_dict(
            data_class=BS5.Scene,
            data=obj,
            config=dacite.Config(
                cast=[Enum, tuple],
            ),
        )

        scene = _fix_data_shape(scene_bs5)

        return scene

    def load_as_bs5(self) -> BS5.Scene:
        raw_data: bytes
        with open(self.scene_url, "rb") as fd:
            raw_data = cast(bytes, fd.read())

        obj = encoder.loads(raw_data)
        if "version" not in obj or not obj["version"].startswith("5."):
            raise Exception(f"Cannot load scene with version {obj['version']}")

        scene_bs5 = dacite.from_dict(
            data_class=BS5.Scene,
            data=obj,
            config=dacite.Config(cast=[Enum, tuple]),
        )

        return scene_bs5



from typing import Sequence, Optional, Tuple

import numpy as np
import numpy.typing as npt
from scale_sensor_fusion_io.models.sensors.camera.camera_sensor import CameraSensor
from scale_sensor_fusion_io.utils.camera_helpers import global_to_local, project_distort


def compute_photocolors(
    positions: npt.NDArray[np.float32],
    cameras: Sequence[CameraSensor],
    points_timestamps: Optional[npt.NDArray[np.uint32]],
    frame_timestamp: Optional[int],
) -> npt.NDArray[np.uint8]:
    """
    Compute photocolors for provided points.

    Photocolors are computed by projecting each point into each camera and
    sampling the color of the pixel that the point projects to.
    """
    # initialize color to shape of point position
    colors = np.full(positions.shape, 255, dtype=np.uint8)

    start_timestamp = (
        int(np.min(points_timestamps))
        if points_timestamps is not None
        else frame_timestamp
    )
    if start_timestamp is None:
        raise ValueError("Must provide either points_timestamps or frame_timestamp")

    # For each camera, find the frame that is closest to the point's timestamp
    for camera in cameras:
        image_data, content_timestamp = camera.get_closest_content_at_timestamp(
            start_timestamp
        )

        intrinsics = camera.intrinsics
        camera_pose = camera.poses.interpolate([content_timestamp]).values[0]

        local_points = global_to_local(
            positions,
            camera_pose,
        )
        try:
            x, y, mask = project_distort(local_points, intrinsics)
        except Exception as e:
            print(f"Failed to project points for camera {camera.id}: {e}")
            continue

        if not hasattr(image_data, 'shape'):
            raise ValueError("image_data must have a shape attribute")
        img_shape: Tuple[int, ...] = tuple(image_data.shape)  # Convert to tuple for safe indexing

        x_arr = np.asarray(x, dtype=np.float64)
        y_arr = np.asarray(y, dtype=np.float64)
        mask_arr = np.asarray(mask, dtype=bool)

        mask = np.logical_and(
            mask_arr,
            np.logical_and(
                np.logical_and(x_arr >= 0, x_arr < img_shape[1]),
                np.logical_and(y_arr >= 0, y_arr < img_shape[0]),
            ),
        )

        index = np.argwhere(mask).reshape(-1)
        
        x_masked = x_arr[mask]
        y_masked = y_arr[mask]

        coords = np.column_stack([x_masked, y_masked])
        pixels = coords.astype(np.int32)

        if not hasattr(colors, 'shape'):
            raise ValueError("colors must have a shape attribute")
        if not hasattr(pixels, 'shape'):
            raise ValueError("pixels must have a shape attribute")
            
        colors[index] = image_data[pixels[:, 1], pixels[:, 0], :]
    return colors



from typing import Union

from .attributes import *
from .box_2d import *
from .cuboid import *
from .event import *
from .labeled_points import *
from .localization_adjustment import *
from .object import *
from .polygon import *

Annotation = Union[
    AttributesAnnotation,
    Box2DAnnotation,
    CuboidAnnotation,
    EventAnnotation,
    LabeledPointsAnnotation,
    LocalizationAdjustmentAnnotation,
    ObjectAnnotation,
    PolygonAnnotation,
]



from .annotations import *
from .common import (
    AnnotationID,
    AnnotationKind,
    FillValue,
    Point3D,
    QuaternionData,
    SensorID,
    SensorKind,
)
from .paths import *
from .Pose import Pose
from .scene import Scene
from .sensors import *



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

import scale_sensor_fusion_io.spec.sfs as SFS

SensorID = SFS.SensorID
AnnotationID = SFS.AnnotationID

PosePath = SFS.PosePath

PointsSensorPoints = SFS.PointsSensorPoints
PointsSensor = SFS.PointsSensor

LidarSensorPoints = SFS.LidarSensorPoints
LidarSensorFrame = SFS.LidarSensorFrame
LidarSensor = SFS.LidarSensor

RadarSensorPoints = SFS.RadarSensorPoints
RadarSensorFrame = SFS.RadarSensorFrame
RadarSensor = SFS.RadarSensor

DistortionModel = SFS.DistortionModel
CameraDistortion = SFS.CameraDistortion

CameraIntrinsics = SFS.CameraIntrinsics

CameraSensorVideo = SFS.CameraSensorVideo
CameraSensorImage = SFS.CameraSensorImage
CameraSensor = SFS.CameraSensor

OdometrySensor = SFS.OdometrySensor

AttributePath = SFS.AttributePath

CuboidPath = SFS.CuboidPath
CuboidActivationPath = SFS.CuboidActivationPath
CuboidProjectionPath = SFS.CuboidProjectionPath


@dataclass
class StaticPath:
    timestamps: List[int]
    values: List[bool]


@dataclass
class CuboidMetadata:
    static_path: Optional[StaticPath] = None


@dataclass
class _CuboidAnnotationBase:
    cuboid_metadata: CuboidMetadata


@dataclass
class CuboidAnnotation(SFS.CuboidAnnotation, _CuboidAnnotationBase):
    pass


AttributesAnnotation = SFS.AttributesAnnotation
AnnotationPath = SFS.AnnotationPath
Box2DAnnotation = SFS.Box2DAnnotation
Polygon2DAnnotation = SFS.Polygon2DAnnotation
Point2DAnnotation = SFS.Point2DAnnotation
PointAnnotation = SFS.PointAnnotation
PolylineAnnotation = SFS.PolylineAnnotation
PolygonAnnotation = SFS.PolygonAnnotation
TopdownPolygonAnnotation = SFS.TopdownPolygonAnnotation
TopdownPointAnnotation = SFS.TopdownPointAnnotation
Polyline2DAnnotation = SFS.Polyline2DAnnotation
EventAnnotation = SFS.EventAnnotation
LabeledPoint = SFS.LabeledPoint
LabeledPointsAnnotation = SFS.LabeledPointsAnnotation


@dataclass
class LinkMetadata:
    anchored: Optional[bool] = False


## NOTE: This is implemented this way to be able to inherit from the SFS dataclasses, which contain defaults
@dataclass
class _LinkAnnotationBase:
    metadata: LinkMetadata


@dataclass
class LinkAnnotation(SFS.LinkAnnotation, _LinkAnnotationBase):
    pass


class CuboidLayerMode(Enum):
    Position = "position"
    PositionRotation = "position-rotation"
    ZLevel = "z-level"
    XY = "XY"
    ICP = "icp"


@dataclass
class LocalizationAdjustmentLayerMetadata:
    layer_type: Literal["base", "cuboid"]
    order: int
    name: str
    cuboids: Optional[List[CuboidPath]] = None
    algorithm: Optional[CuboidLayerMode] = None


## NOTE: This is implemented this way to be able to inherit from the SFS dataclasses, which contain defaults
@dataclass
class _LocalizationAdjustmentAnnotationBase:
    layer_metadata: LocalizationAdjustmentLayerMetadata


@dataclass
class LocalizationAdjustmentAnnotation(
    SFS.LocalizationAdjustmentAnnotation, _LocalizationAdjustmentAnnotationBase
):
    pass


ObjectAnnotation = SFS.ObjectAnnotation

Sensor = Union[CameraSensor, LidarSensor, RadarSensor, OdometrySensor, PointsSensor]


Annotation = Union[
    CuboidAnnotation,
    AttributesAnnotation,
    Box2DAnnotation,
    Point2DAnnotation,
    Polyline2DAnnotation,
    Polygon2DAnnotation,
    TopdownPolygonAnnotation,
    TopdownPointAnnotation,
    PolygonAnnotation,
    PolylineAnnotation,
    PointAnnotation,
    LinkAnnotation,
    LabeledPointsAnnotation,
    LocalizationAdjustmentAnnotation,
    ObjectAnnotation,
]


@dataclass
class Scene:
    version: Literal["5.1", "1.0"] = "5.1"
    sensors: Optional[List[Sensor]] = None
    annotations: Optional[List[Annotation]] = None
    attributes: Optional[List[AttributePath]] = None
    time_offset: Optional[int] = None
    time_unit: Optional[Literal["microseconds", "nanoseconds"]] = "microseconds"
    metadata: Optional[Dict[str, Any]] = None



from .attribute_path import AttributePath
from .cuboid_path import CuboidPath
from .spacial_path import SpacialPath, SpacialPathT
from .pose_path import PosePath
from .transformable_path import ITransformablePath



from dataclasses import dataclass
from typing import List, Literal, Optional, Union

from ..common import AnnotationID, AnnotationKind, SensorID
from ..paths import AttributePath


@dataclass
class EventAnnotation:
    id: AnnotationID
    start: int
    type: Literal[AnnotationKind.Event] = AnnotationKind.Event
    parent_id: Optional[AnnotationID] = None
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None
    duration: Optional[int] = None
    sensor_id: Optional[SensorID] = None



from dataclasses import dataclass
from typing import Literal, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt

from ...common import SensorID, SensorKind
from ...paths import PosePath


@dataclass
class RadarSensorPoints:
    positions: npt.NDArray[np.float32]
    directions: Optional[npt.NDArray[np.float32]] = None
    lengths: Optional[npt.NDArray[np.float32]] = None
    timestamps: Optional[Union[npt.NDArray[np.uint32], npt.NDArray[np.uint64]]] = None


@dataclass
class RadarSensorFrame:
    timestamp: int
    points: RadarSensorPoints


@dataclass
class RadarSensor:
    id: SensorID
    poses: PosePath
    frames: Sequence[RadarSensorFrame]
    type: Literal[SensorKind.Radar] = SensorKind.Radar
    coordinates: Literal["ego", "world"] = "world"
    parent_id: Optional[SensorID] = None



"""Models shared across other models."""
from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import Any, Literal, Optional, Type, Union

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation

AnnotationID = Union[str, int]
FillValue = Union[Literal["nearest"], Literal["identity"], Literal["extrapolate"]]
SensorID = Union[str, int]


class AnnotationKind(Enum):
    Attributes = "attributes"
    Box2D = "box_2d"
    Cuboid = "cuboid"
    Event = "event"
    LabeledPoints = "labeled_points"
    PointSegmentation = "point_segmentation"
    LocalizationAdjustment = "localization_adjustment"
    Object = "object"
    Polygon = "polygon"


@dataclass
class Point3D:
    """A point in 3D space."""

    x: float
    y: float
    z: float

    def copy(self) -> Point3D:
        """Creates a copy of this object.

        Returns
        -------
        Point3D
            A clone of this object.
        """
        return Point3D(x=self.x, y=self.y, z=self.z)

    @staticmethod
    def identity() -> Point3D:
        """Represents no translation.

        Returns
        -------
        Point3D
            (0, 0, 0)
        """
        return Point3D(x=0.0, y=0.0, z=0.0)

    def __array__(
        self, dtype: Optional[Union[np.dtype, str, Type]] = None
    ) -> npt.NDArray[np.float64]:
        return np.array([self.x, self.y, self.z], dtype=dtype)

    def __eq__(self, other: Any) -> bool:
        if (
            hasattr(other, "x")
            and hasattr(other, "y")
            and hasattr(other, "z")
            and self.x == other.x
            and self.y == other.y
            and self.z == other.z
        ):
            return True
        us = self.__array__()
        other = np.asarray(other)
        return np.prod(us.shape) == np.prod(other.shape) and np.allclose(
            us, other.reshape(us.shape)
        )

    def __repr__(self) -> str:
        return f"(x={float(self.x):.3} y={float(self.y):.3} z={float(self.z):.3})"


@dataclass
class QuaternionData:
    """A Quaternion representing rotation in 3D space."""

    x: float
    y: float
    z: float
    w: float

    def copy(self) -> QuaternionData:
        """Creates a copy of this object.

        Returns
        -------
        QuaternionData
            Clone of this object.
        """
        return QuaternionData(x=self.x, y=self.y, z=self.z, w=self.w)

    def as_euler(
        self, seq: str = "XYZ", degrees: bool = False
    ) -> npt.NDArray[np.float64]:
        """Get this object as Euler angles.

        Parameters
        ----------
        seq : str, optional
            3 characters belonging to the set {'X', 'Y', 'Z'} for intrinsic rotations,
            or {'x', 'y', 'z'} for extrinsic rotations [1]. Adjacent axes cannot be the same.
            Extrinsic and intrinsic rotations cannot be mixed in one function call.
            By default "XYZ"
        degrees : bool, optional
            Whether angles are in degrees, by default False

        Returns
        -------
        npt.NDArray
            Vector of Euler angles.
        """
        return np.asarray(self.as_rotation().as_euler(seq, degrees=degrees))

    @staticmethod
    def from_euler(
        angles: npt.ArrayLike,
        seq: str = "XYZ",
        degrees: bool = False,
    ) -> QuaternionData:
        """Create a transform from Euler angles.

        Parameters
        ----------
        angles : ArrayLike[float]
            Euler angles
        seq : str, optional
            3 characters belonging to the set {'X', 'Y', 'Z'} for intrinsic rotations,
            or {'x', 'y', 'z'} for extrinsic rotations [1]. Adjacent axes cannot be the same.
            Extrinsic and intrinsic rotations cannot be mixed in one function call.
            By default "XYZ"
        degrees : bool, optional
            Whether angles are in degrees, by default False

        Returns
        -------
        QuaternionData:
            The angles as a Quaternion.
        """
        return QuaternionData(*Rotation.from_euler(seq, angles, degrees).as_quat())

    def as_matrix(self) -> npt.NDArray[np.float64]:
        """Creates a rotation matrix.

        Returns
        -------
        npt.NDArray[np.float64]:
            A 3x3 rotation matrix.
            ```py
            [
                [ r00, r01, r02],
                [ r10, r11, r12],
                [ r20, r21, r22],
            ]
            ```
        """
        return np.asarray(Rotation.from_quat(self).as_matrix(), dtype=np.float64)

    @staticmethod
    def from_matrix(rotation_matrix: npt.ArrayLike) -> QuaternionData:
        """
        Creates a QuaternionData object from a rotation matrix.

        Parameters
        ----------
        rotation_matrix : npt.ArrayLike
            An object representing a 3x3 rotation matrix.
            ```py
            [
                [ r00, r01, r02],
                [ r10, r11, r12],
                [ r20, r21, r22],
            ]
            ```

        Returns
        -------
        QuaternionData
            A QuaternionData object corresponding to the provided rotation.
        """
        return QuaternionData(
            *Rotation.from_matrix(
                np.asarray(rotation_matrix, dtype=np.float64).reshape((3, 3))
            ).as_quat()
        )

    def as_rotation(self) -> Rotation:
        """Gets this object as a scipy Rotation object.

        Returns
        -------
        Rotation
            Equivalent Rotation object.
        """
        return Rotation.from_quat(self)

    @staticmethod
    def from_rotation(rotation: Rotation) -> QuaternionData:
        """Creates a QuaternionData object from scipy Rotation.

        Parameters
        ----------
        rotation : Rotation
            The rotation to copy.

        Returns
        -------
        QuaternionData
            Equivalent QuaternionData object.
        """
        return QuaternionData(*rotation.as_quat())

    @staticmethod
    def identity() -> QuaternionData:
        """Represents no rotation.

        Returns
        -------
        QuaternionData
            (x=0, y=0, z=0, w=1)
        """
        return QuaternionData(x=0.0, y=0.0, z=0.0, w=1.0)

    def __array__(
        self, dtype: Optional[Union[np.dtype, str, Type]] = None
    ) -> npt.NDArray[np.float64]:
        return np.array([self.x, self.y, self.z, self.w], dtype=dtype)

    def __eq__(self, other: Any) -> bool:
        if (
            hasattr(other, "x")
            and hasattr(other, "y")
            and hasattr(other, "z")
            and hasattr(other, "w")
            and self.x == other.x
            and self.y == other.y
            and self.z == other.z
            and self.w == other.w
        ):
            return True
        us = self.__array__()
        other = np.asarray(other)
        return np.prod(us.shape) == np.prod(other.shape) and np.allclose(
            us, other.reshape(us.shape)
        )

    def __repr__(self) -> str:
        return f"(x={float(self.x):.3} y={float(self.y):.3} z={float(self.z):.3} w={float(self.w):.3})"


class SensorKind(Enum):
    Camera = "camera"
    Lidar = "lidar"
    Radar = "radar"
    Points = "points"
    Odometer = "odometer"



from dataclasses import asdict, dataclass, is_dataclass
from enum import Enum
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Literal,
    Mapping,
    Protocol,
    Type,
    Union,
)

from typing_extensions import TypeAlias


class DistortionModel(str, Enum):
    FISHEYE = "fisheye"
    MOD_KANNALA = "mod_kannala"
    CYLINDRICAL = "cylindrical"
    OMNIDIRECTIONAL = "omnidirectional"
    BROWN_CONRADY = "brown_conrady"
    EQUIRECTANGULAR = "equirectangular"
    MOD_EQUI_FISH = "mod_equi_fish"
    FISHEYE_RAD_TAN_THIN_PRISM = "fisheye_rad_tan_prism"
    FISHEYE_RADIAL_CUSTOM = "fisheye_radial_custom"
    FTHETA = "ftheta"


@dataclass
class DistortionParamsBase:
    model: DistortionModel


@dataclass
class BrownConradyParams(DistortionParamsBase):
    model: Literal[DistortionModel.BROWN_CONRADY] = DistortionModel.BROWN_CONRADY
    k1: float = 0
    k2: float = 0
    p1: float = 0
    p2: float = 0
    k3: float = 0
    k4: float = 0
    k5: float = 0
    k6: float = 0
    lx: float = 0
    ly: float = 0


@dataclass
class FisheyeParams(DistortionParamsBase):
    model: Literal[DistortionModel.FISHEYE] = DistortionModel.FISHEYE
    k1: float = 0
    k2: float = 0
    k3: float = 0
    k4: float = 0


@dataclass
class AppleFisheyeParams(DistortionParamsBase):
    model: Literal[DistortionModel.MOD_EQUI_FISH] = DistortionModel.MOD_EQUI_FISH
    k1: float = 0
    k2: float = 0
    k3: float = 0
    k4: float = 0


@dataclass
class AppleKannalaParams(DistortionParamsBase):
    model: Literal[DistortionModel.MOD_KANNALA] = DistortionModel.MOD_KANNALA
    k1: float = 0
    k2: float = 0
    k3: float = 0
    k4: float = 0


@dataclass
class CylindricalParams(DistortionParamsBase):
    model: Literal[DistortionModel.CYLINDRICAL] = DistortionModel.CYLINDRICAL

@dataclass
class EquirectangularParams(DistortionParamsBase):
    model: Literal[DistortionModel.EQUIRECTANGULAR] = DistortionModel.EQUIRECTANGULAR


@dataclass
class OmnidirectionalParams(DistortionParamsBase):
    model: Literal[DistortionModel.OMNIDIRECTIONAL] = DistortionModel.OMNIDIRECTIONAL
    k1: float = 0
    k2: float = 0
    k3: float = 0
    k4: float = 0
    p1: float = 0
    p2: float = 0
    xi: float = 0
    lx: float = 0
    ly: float = 0


@dataclass
class FisheyeRadTanThinPrismParams(DistortionParamsBase):
    model: Literal[
        DistortionModel.FISHEYE_RAD_TAN_THIN_PRISM
    ] = DistortionModel.FISHEYE_RAD_TAN_THIN_PRISM
    k1: float = 0
    k2: float = 0
    k3: float = 0
    k4: float = 0
    k5: float = 0
    k6: float = 0
    p1: float = 0
    p2: float = 0
    s1: float = 0
    s2: float = 0
    s3: float = 0
    s4: float = 0


@dataclass
class FisheyeRadialCustom(DistortionParamsBase):
    model: Literal[
        DistortionModel.FISHEYE_RADIAL_CUSTOM
    ] = DistortionModel.FISHEYE_RADIAL_CUSTOM
    k1: float = 0
    k2: float = 0
    k3: float = 0
    k4: float = 0
    k5: float = 0
    k6: float = 0
    p1: float = 0
    p2: float = 0
    dcx: float = 0
    dcy: float = 0


@dataclass
class FthetaParams(DistortionParamsBase):
    model: Literal[DistortionModel.FTHETA] = DistortionModel.FTHETA
    k1: float = 0
    k2: float = 0
    k3: float = 0
    k4: float = 0
    k5: float = 0


DistortionParams: TypeAlias = Union[
    BrownConradyParams,
    FisheyeParams,
    AppleFisheyeParams,
    AppleKannalaParams,
    CylindricalParams,
    EquirectangularParams,
    OmnidirectionalParams,
    FisheyeRadTanThinPrismParams,
    FisheyeRadialCustom,
    FthetaParams,
]

DISTORTION_PARAMETERS = {
    params
    for dist_models in DistortionParams.__args__  # type: ignore
    for params in dist_models.__dataclass_fields__.keys()
}


class DataClassProtocol(Protocol):
    __dataclass_fields__: ClassVar[dict]


def get_params_from_model(model: DistortionModel) -> Type[DistortionParams]:
    """
    Returns the dataclass for the given distortion model
    """
    supported_models: List[Type[DistortionParams]] = [
        BrownConradyParams,
        FisheyeParams,
        AppleFisheyeParams,
        AppleKannalaParams,
        CylindricalParams,
        EquirectangularParams,
        OmnidirectionalParams,
        FisheyeRadTanThinPrismParams,
        FisheyeRadialCustom,
        FthetaParams,
    ]

    return next(i for i in supported_models if i.model == model)


def extract_distortion_params(
    distortion: Union[DataClassProtocol, Mapping, dict]
) -> DistortionParams:
    """
    Generic helper to extract distortion params from untyped inputs.
    """
    distortion_fields: Dict[str, Any] = {}
    if isinstance(distortion, dict):
        distortion_fields = distortion
    elif isinstance(distortion, Mapping):
        distortion_fields = {k: v for k, v in distortion.items()}
    elif is_dataclass(distortion):
        distortion_fields = asdict(distortion)
    else:
        distortion_fields = distortion.__dict__

    model: DistortionModel = (
        distortion_fields.get("model")
        or distortion_fields.get("camera_model")
        or DistortionModel.BROWN_CONRADY
    )

    dataclass = get_params_from_model(model)
    # construct param dict for dataclass
    d = {}
    for key in dataclass.__dataclass_fields__:
        if (
            key in distortion_fields
            and key != "model"
            and distortion_fields[key] is not None
        ):
            d[key] = distortion_fields[key]
    return dataclass(**d)


# Define CameraDistortion dataclass
@dataclass
class CameraDistortion:
    """
    NOTE: model is here as helper attribute since it is already included in params
    """

    model: DistortionModel
    params: DistortionParams

    @staticmethod
    def from_values(model: str, values: List[float]) -> "CameraDistortion":
        """
        Creates a CameraDistortion IR from a model name and a list of values.

        NOTE: IMPORTANT: The values MUST be in the same order as the dataclass
        """
        # This will throw a ValueError if model is not a valid DistortionModel
        model_ = DistortionModel(model)
        ParamsClass = get_params_from_model(model_)

        # NOTE: This is a hack because neither of the following will work with
        # our current version of mypy
        #
        # 1. ParamsClass(model_, *values): Complains that model_ is not of type
        #    Literal[DistortionModel]
        # 2. params_dict = {...} (without the type hint): Complains that the
        #    Dict type is not compatible with the dataclass init. See:
        #    https://github.com/python/mypy/issues/5382
        # Therefore, we must type the dict as Dict[str, Any] to satisfy mypy
        params_dict: Dict = {
            k: v
            for k, v in zip(
                [k for k in ParamsClass.__dataclass_fields__.keys() if k != "model"],
                values,
            )
        }
        params = ParamsClass(**params_dict)
        return CameraDistortion(
            model=model_,
            params=params,
        )

    @staticmethod
    def from_dict(distortion_dict: Dict[str, Any]) -> "CameraDistortion":
        """
        Creates a CameraDistortion IR from a dict
        """
        model = DistortionModel(distortion_dict["model"])
        params = extract_distortion_params(distortion_dict)
        return CameraDistortion(
            model=model,
            params=params,
        )



from dataclasses import dataclass
from typing import List, Literal, Optional

from ..common import AnnotationID, AnnotationKind, SensorID
from ..paths import AttributePath


@dataclass
class PolygonPath:
    timestamps: List[int]
    values: List[List[float]]  # x_0, y_0, x_1, y_1, ..., x_n, y_n


@dataclass
class PolygonAnnotation:
    id: AnnotationID
    sensor_id: SensorID
    path: PolygonPath
    type: Literal[AnnotationKind.Polygon] = AnnotationKind.Polygon
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None



from typing import (
    List,
    Optional,
    TypeVar,
    Callable,
)
from .error import ErrorDetails

import scale_sensor_fusion_io.validation.dacite_internal as dacite

T = TypeVar("T")


def convert_error(e: dacite.DaciteFieldError) -> ErrorDetails:
    return ErrorDetails(
        path=e.field_path.split(".") if e.field_path else [], errors=[str(e)]  # type: ignore
    )


def handle_dacite(
    fn: Callable[[], T], error_details: List[ErrorDetails]
) -> Optional[T]:
    """Handle dacite errors and aggregate errors"""
    try:
        return fn()
    except dacite.AggregatedError as e:
        error_details.extend([convert_error(err) for err in e.errors])
    except dacite.DaciteFieldError as e:
        error_details.append(convert_error(e))
    return None


def is_strictly_increasing(ts: List[int]) -> bool:
    """Check if a list of timestamps is strictly increasing"""
    return all(ts1 < ts2 for ts1, ts2 in zip(ts[:-1], ts[1:]))



import os
from dataclasses import dataclass
from typing import List, Literal, Optional, Tuple
import tempfile

import numpy as np
import numpy.typing as npt
from scale_sensor_fusion_io.models.sensors.camera import (
    CameraIntrinsics,
)
from scale_sensor_fusion_io.utils.video_helpers import VideoReader

from ...common import SensorID, SensorKind
from ...paths import PosePath


@dataclass
class CameraSensorVideo:
    timestamps: List[int]
    content: npt.NDArray[np.uint8]
    fps: float


@dataclass
class CameraSensorImage:
    timestamp: int
    content: npt.NDArray[np.uint8]


@dataclass
class CameraSensor:
    id: SensorID
    poses: PosePath
    intrinsics: CameraIntrinsics
    video: Optional[CameraSensorVideo] = None
    images: Optional[List[CameraSensorImage]] = None
    type: Literal[SensorKind.Camera] = SensorKind.Camera
    parent_id: Optional[SensorID] = None
    coordinates: Literal["ego", "world"] = "world"

    video_loader: Optional[VideoReader] = None

    def __init__(
        self,
        id: SensorID,
        poses: PosePath,
        intrinsics: CameraIntrinsics,
        video: Optional[CameraSensorVideo] = None,
        images: Optional[List[CameraSensorImage]] = None,
        type: Literal[SensorKind.Camera] = SensorKind.Camera,
        parent_id: Optional[SensorID] = None,
    ) -> None:
        self.id = id
        self.poses = poses
        self.intrinsics = intrinsics
        self.video = video
        self.images = images
        self.type = type
        self.parent_id = parent_id

        if video is not None:
            self.video_loader: Optional[VideoReader] = VideoReader(
                video=video.content.tobytes()
            )

    def get_closest_content_at_timestamp(
        self, timestamp: int
    ) -> Tuple[npt.NDArray[np.uint8], int]:
        content_timestamps: List[int]
        if self.video:
            content_timestamps = self.video.timestamps
        elif self.images:
            content_timestamps = [image.timestamp for image in self.images]
        else:
            raise Exception("Missing camera content")

        # Find the closest timestamp to the provided timestamp
        closest_frame = int(np.argmin(np.abs(np.array(content_timestamps) - timestamp)))

        return (
            self.get_content_at_frame(closest_frame),
            content_timestamps[closest_frame],
        )

    def get_content_at_frame(self, frame: int) -> npt.NDArray[np.uint8]:
        if self.video:
            if not self.video_loader:
                self.video_loader: Optional[VideoReader] = VideoReader(
                    video=self.video.content.tobytes()
                )
                if not self.video_loader:
                    raise Exception("Missing video loader")
            return self.video_loader.load(frame)
        elif self.images:
            return self.images[frame].content
        else:
            raise Exception("Missing camera content")



from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy.spatial.transform import Rotation

from typing import Optional, Union
from numpy.typing import ArrayLike

from .spacial_path import SpacialPath, SpacialInterpolationData


class CuboidPath(SpacialPath):
    """CuboidPath class representing a list of cuboids at given timestamps, extending pandas DataFrame."""

    POSITION = ["x", "y", "z"]
    ROTATION = ["yaw", "pitch", "roll"]
    DIMENSIONS = ["dx", "dy", "dz"]
    COLUMNS = POSITION + ROTATION + DIMENSIONS

    def __init__(self, data: Union[ArrayLike, pd.DataFrame], index: ArrayLike = None):
        """Initializes the CuboidPath object.

        Args:
            data (Union[ArrayLike, pd.DataFrame]): An array or DataFrame of cuboids with shape (N, 9), where N is the number of cuboids.
            index (ArrayLike, optional): An array-like object representing the index for the CuboidPath DataFrame. Defaults to None.
        """
        super().__init__(
            data=data, index=index, columns=CuboidPath.COLUMNS, dtype=float
        )

    def copy(self):
        """Creates a copy of the current CuboidPath object.

        Returns:
            CuboidPath: A new CuboidPath object with copied data and index.
        """
        return CuboidPath(self.values.copy(), index=self.index)

    @property
    def positions(self):
        return self[CuboidPath.POSITION].values

    @property
    def rotations(self):
        return self[CuboidPath.ROTATION].values

    @property
    def dimensions(self):
        return self[CuboidPath.DIMENSIONS].values

    @classmethod
    def from_csv(cls, file: str):
        """Creates a CuboidPath object from a CSV file.

        Args:
            file (str): The path to the CSV file.

        Returns:
            CuboidPath: A CuboidPath object with data read from the CSV file.
        """
        return CuboidPath(pd.read_csv(file, index_col=0))

    @classmethod
    def identity(cls, n: int = 1, index: ArrayLike = None):
        """Create a CuboidPath object with identity cuboids.

        Args:
            n (int, optional): The number of identity cuboids. Defaults to 1.
            index (ArrayLike, optional): An array-like object representing the index for the CuboidPath DataFrame. Defaults to None.

        Returns:
            CuboidPath: A CuboidPath object with identity cuboids.
        """
        identity_cuboid = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1])
        return CuboidPath(np.tile(identity_cuboid, n).reshape((n, 9)), index=index)

    @classmethod
    def from_matrix(
        cls, matrix: ArrayLike, index: Optional[ArrayLike] = None
    ) -> CuboidPath:
        """Creates a CuboidPath object from transformation matrices.

        Args:
            matrix (ArrayLike): A 3D array-like object of transformation matrices with shape (N, 4, 4), where N is the number of matrices.
            index (ArrayLike, optional): An array-like object representing the index for the CuboidPath DataFrame. Defaults to None.

        Returns:
            CuboidPath: A CuboidPath object with cuboids represented by the given transformation matrices.
        """
        matrix = np.asarray(matrix).reshape((-1, 4, 4))
        positions = matrix[:, :3, 3]
        scales = np.linalg.norm(matrix[:, :3, :3], axis=1)
        rotations = Rotation.from_matrix(
            matrix[:, :3, :3] / scales[:, np.newaxis, :]
        ).as_euler("zyx", degrees=True)

        cuboids = np.hstack([positions, rotations, scales])
        return CuboidPath(cuboids, index=index)

    def as_matrix(self):
        """Convert the CuboidPath object to transformation matrices.

        Returns:
            np.ndarray: A 3D array of transformation matrices with shape (N, 4, 4), where N is the number of cuboids.
        """
        positions = self.loc[:, CuboidPath.POSITION].values
        rotations = self.loc[:, CuboidPath.ROTATION].values
        dimensions = self.loc[:, CuboidPath.DIMENSIONS].values

        matrix = np.tile(np.eye(4), (len(self), 1, 1))

        rotation_matrices = Rotation.from_euler(
            "zyx", rotations, degrees=True
        ).as_matrix()
        scale_matrices = np.array([np.diag(s) for s in dimensions])

        # Combine scale and rotation
        transform_matrices = np.matmul(rotation_matrices, scale_matrices)

        matrix[:, :3, :3] = transform_matrices
        matrix[:, :3, 3] = positions

        return matrix

    # InterpolateablePath
    @classmethod
    def from_interpolation_data(
        cls, value: SpacialInterpolationData, index: npt.NDArray
    ) -> CuboidPath:
        """Creates a CuboidPath from interpolated values.

        Parameters
        ----------
        values : SpacialInterpolationData
            Position, rotation, and dimension values.
        index : npt.NDArray
            Timestamps

        Returns
        -------
        CuboidPath
            Result of interpolation.
        """
        assert value.linear is not None, "Value has no linear component"
        return CuboidPath(value.linear, index=index)

    def get_interpolation_data(self) -> SpacialInterpolationData:
        """Gets the values of this object for interpolation. Rotation interpolation uses lerp on
        Euler angles to match LidarLite. As long as only 1 of 3 Euler angles is large, this is a
        good approximation if slerp.

        Returns
        -------
        SpacialInterpolationData
            Position, rotation, and dimension values.
        """
        return SpacialInterpolationData(linear=self.values)



from dataclasses import dataclass
from typing import Literal, Optional

from ...common import SensorID, SensorKind
from ...paths import PosePath


# Define Odometry dataclass
@dataclass
class OdometrySensor:
    id: SensorID
    poses: PosePath
    parent_id: Optional[SensorID] = None
    type: Literal[SensorKind.Odometer] = SensorKind.Odometer



from dataclasses import dataclass
from typing import List, Literal, Optional, Union, Sequence

import numpy as np
import numpy.typing as npt

from ...common import SensorID, SensorKind
from ...paths import PosePath


# Define LidarSensorPoints dataclass
@dataclass
class LidarSensorPoints:
    positions: npt.NDArray[np.float32]
    colors: Optional[npt.NDArray[np.uint8]] = None
    intensities: Optional[npt.NDArray[np.uint8]] = None
    timestamps: Optional[Union[npt.NDArray[np.uint32], npt.NDArray[np.uint64]]] = None


# Define LidarSensorFrame dataclass
@dataclass
class LidarSensorFrame:
    timestamp: int
    points: LidarSensorPoints


# Define LidarSensor dataclass
@dataclass
class LidarSensor:
    id: SensorID
    poses: PosePath
    frames: Sequence[LidarSensorFrame]
    parent_id: Optional[SensorID] = None
    coordinates: Literal["ego", "world"] = "world"
    type: Literal[SensorKind.Lidar] = SensorKind.Lidar



import copy
from dataclasses import is_dataclass, dataclass, Field
from itertools import zip_longest
from typing import TypeVar, Type, Optional, Mapping, Any, Union, List
from typing_extensions import TypeAlias

from .config import Config
from .data import Data, DictData
from .dataclasses import (
    get_default_value_for_field,
    create_instance,
    DefaultValueNotFoundError,
    get_fields,
)
from .exceptions import (
    ForwardReferenceError,
    WrongTypeError,
    DaciteError,
    UnionMatchError,
    MissingValueError,
    DaciteFieldError,
    UnexpectedDataError,
    StrictUnionMatchError,
    AggregatedError,
)
from .types import (
    is_instance,
    is_generic_collection,
    is_union,
    extract_generic,
    is_optional,
    transform_value,
    extract_origin_collection,
    is_init_var,
    extract_init_var,
    is_set,
    get_data_class_hints,
)

T = TypeVar("T")


SKIP_FLAG = "SKIP_FLAG"


def cast_field(data: Data, field: Field, config: Config) -> Any:
    if field.name in data:
        try:
            field_data = data[field.name]
            transformed_value = transform_value(
                type_hooks=config.type_hooks,
                cast=config.cast,
                target_type=field.type,
                value=field_data,
            )
            value = _build_value(
                type_=field.type, data=transformed_value, config=config
            )

        except AggregatedError as e:
            for err in e.errors:
                err.update_path(field.name)
            raise

        except DaciteFieldError as error:
            error.update_path(field.name)
            raise
        if config.check_types and not is_instance(value, field.type):
            raise WrongTypeError(
                field_path=field.name, field_type=field.type, value=value
            )

        return value
    elif not field.init:
        # If the non-init field isn't in the dict, let the dataclass handle default, to ensure
        # we won't get errors in the case of frozen dataclasses, as issue #195 highlights.
        return SKIP_FLAG
    else:
        try:
            return get_default_value_for_field(
                field,
                allow_missing_fields_as_none=config.allow_missing_fields_as_none,
            )
        except DefaultValueNotFoundError:
            raise MissingValueError(field.name)


def from_dict(data_class: Type[T], data: Data, config: Optional[Config] = None) -> T:
    """Create a data class instance from a dictionary.

    :param data_class: a data class type
    :param data: a dictionary of a input data
    :param config: a configuration of the creation process
    :return: an instance of a data class
    """
    init_values: DictData = {}
    post_init_values: DictData = {}
    config = config or Config()
    errors = []
    try:
        data_class_hints = get_data_class_hints(
            data_class, globalns=config.forward_references
        )
    except NameError as error:
        raise ForwardReferenceError(str(error))
    data_class_fields = get_fields(data_class)
    if config.strict:
        extra_fields = set(data.keys()) - {f.name for f in data_class_fields}
        if extra_fields:
            raise UnexpectedDataError(keys=extra_fields)
    for field in data_class_fields:
        field = copy.copy(field)
        field.type = data_class_hints[field.name]
        # Wrap try catch so we can capture multiple failed fields at a time
        try:
            value = cast_field(data, field, config)
            if type(value) is str and value == SKIP_FLAG:
                continue

            if field.init:
                init_values[field.name] = value
            else:
                post_init_values[field.name] = value
        except AggregatedError as e:
            errors.extend(e.errors)

        except DaciteFieldError as e:
            errors.append(e)

    if len(errors) > 0:
        raise AggregatedError(errors)

    return create_instance(
        data_class=data_class,
        init_values=init_values,
        post_init_values=post_init_values,
    )


def _build_value(type_: Type, data: Any, config: Config) -> Any:
    if is_init_var(type_):
        type_ = extract_init_var(type_)
    if is_union(type_):
        return _build_value_for_union(union=type_, data=data, config=config)
    elif is_generic_collection(type_):
        origin = extract_origin_collection(type_)
        if is_instance(data, origin):
            return _build_value_for_collection(
                collection=type_, data=data, config=config
            )
        if is_set(origin):
            return origin(
                _build_value(
                    type_=extract_generic(type_)[0], data=single_val, config=config
                )
                for single_val in data
            )
    elif is_dataclass(type_) and is_instance(data, Data):
        if hasattr(type_, "from_dict"):
            return type_.from_dict(data=data, config=config)
        return from_dict(data_class=type_, data=data, config=config)

    return data


def _build_value_for_union(union: Type, data: Any, config: Config) -> Any:
    types = extract_generic(union)
    if is_optional(union) and len(types) == 2:
        return _build_value(type_=types[0], data=data, config=config)
    union_matches = {}
    for inner_type in types:
        try:
            # noinspection PyBroadException
            try:
                data = transform_value(
                    type_hooks=config.type_hooks,
                    cast=config.cast,
                    target_type=inner_type,
                    value=data,
                )
            except Exception as e:  # pylint: disable=broad-except
                continue
            value = _build_value(type_=inner_type, data=data, config=config)
            if is_instance(value, inner_type):
                if config.strict_unions_match:
                    union_matches[inner_type] = value
                else:
                    return value
        except DaciteError as e:
            print(e)
            pass
    if config.strict_unions_match:
        if len(union_matches) > 1:
            raise StrictUnionMatchError(union_matches)
        return union_matches.popitem()[1]
    if not config.check_types:
        return data
    raise UnionMatchError(field_type=union, value=data)


def _build_value_for_collection(collection: Type, data: Any, config: Config) -> Any:
    data_type = data.__class__
    if is_instance(data, Mapping):
        item_type = extract_generic(collection, defaults=(Any, Any))[1]
        return data_type(
            (key, _build_value(type_=item_type, data=value, config=config))
            for key, value in data.items()
        )
    elif is_instance(data, tuple):
        if not data:
            return data_type()
        types = extract_generic(collection)
        if len(types) == 2 and types[1] == Ellipsis:
            return data_type(
                _build_value(type_=types[0], data=item, config=config) for item in data
            )
        return data_type(
            _build_value(type_=type_, data=item, config=config)
            for item, type_ in zip_longest(data, types)
        )
    item_type = extract_generic(collection, defaults=(Any,))[0]
    return data_type(
        _build_value(type_=item_type, data=item, config=config) for item in data
    )



from typing import List, Protocol, Tuple, TypeVar, Union, cast
from typing_extensions import assert_never

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation
from scale_sensor_fusion_io.models.sensors.camera import (
    AppleFisheyeParams,
    AppleKannalaParams,
    BrownConradyParams,
    DistortionModel,
    FisheyeParams,
    OmnidirectionalParams,
)
from scale_sensor_fusion_io.models.common import Point3D, QuaternionData
from scale_sensor_fusion_io.models.sensors.camera.camera_intrinsics import (
    CameraIntrinsics,
)


def global_to_local(
    pts: npt.NDArray[np.float64],
    cam_pose: npt.NDArray[np.float64],  # [x,y,z,qx,qy,qz,qw]
) -> npt.NDArray[np.float64]:
    local_pts = pts - cam_pose[0:3]
    rotation_matrix: npt.NDArray[np.float64] = Rotation.from_quat(
        cam_pose[3:7]
    ).as_matrix()
    return cast(npt.NDArray[np.float64], local_pts @ rotation_matrix)


def _smallest_pos_root(coeffs: npt.ArrayLike) -> float:
    roots = np.roots(coeffs)
    pos_roots: List[float] = sorted(roots[roots > 0])
    return pos_roots[0] if pos_roots else float("inf")


def _find_rsq_thresh(params: Union[BrownConradyParams, OmnidirectionalParams]) -> float:
    return _smallest_pos_root([7 * params.k3, 5 * params.k2, 3 * params.k1, 1])


def _find_theta_thresh(params: AppleFisheyeParams) -> float:
    return _smallest_pos_root([5 * params.k4, 4 * params.k3, 3 * params.k2, 0, 1])


def _find_theta_thresh_kannala(params: AppleKannalaParams) -> float:
    return np.pi / 2


def _find_theta2_thresh(params: FisheyeParams) -> float:
    return _smallest_pos_root(
        [9 * params.k4, 7 * params.k3, 5 * params.k2, 3 * params.k1, 1]
    )


def project_distort(
    pts: npt.NDArray, intrinsics: CameraIntrinsics
) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray[np.bool_]]:
    """
    given points in 3d relative to the camera, and a set of camera params
    returns a 3-tuple of distorted image x-coords, y-coords,
    and mask of whether or not the point was likely valid
    (ie within region where distorted radius varies monotonically with undistorted radius, and the point was in front of the camera plane)
    """
    epsilon = 1e-5

    p = intrinsics
    d = intrinsics.distortion.params if intrinsics.distortion else None

    camera_in_front = pts[:, 2] > 0

    if d and d.model is DistortionModel.OMNIDIRECTIONAL and d.xi:
        pts = pts / np.linalg.norm(pts, axis=1, keepdims=True)
        pts[:, 2] += d.xi

    # get x, y in homogeneous coordinates
    x = pts[:, 0] / pts[:, 2]
    y = pts[:, 1] / pts[:, 2]

    if (
        p.skew
    ):  # apply skew now so that we don't have to deal with it during distortion calculation
        x += y * p.skew / p.fx

    if not d or d.model is DistortionModel.CYLINDRICAL:
        # no distortion
        u = p.cx + p.fx * x
        v = p.cy + p.fy * y
        return u, v, camera_in_front

    if d.model is DistortionModel.EQUIRECTANGULAR:
        # Convert to spherical coordinates
        r = np.hypot(x, y)
        theta = np.arctan2(y, x)  # azimuthal angle
        phi = np.arccos(1 / r)  # polar angle

        # Equirectangular projection
        u = theta * p.fx + p.cx
        v = phi * p.fy + p.cy

        return u, v, camera_in_front

    if d.model is DistortionModel.MOD_EQUI_FISH:
        rho = np.hypot(x, y)
        theta = np.arctan(rho)
        m = np.where(rho < epsilon, 1, theta / rho)
        rad_dist_fac = np.polyval([d.k4, d.k3, d.k2, 0, 1], theta)
        u_d = p.fx * m * x * rad_dist_fac + p.cx
        v_d = p.fy * m * y * rad_dist_fac + p.cy
        return u_d, v_d, (theta < _find_theta_thresh(d)) & camera_in_front

    if d.model is DistortionModel.MOD_KANNALA:
        rho = np.hypot(x, y)
        theta = np.arctan(rho)
        ftheta = np.arctan(theta)
        ftheta2 = ftheta * ftheta
        r = np.polyval([d.k4, d.k3, d.k2, d.k1, 1], ftheta2) * ftheta
        rad_dist_fac = np.where(r < epsilon, 1, np.tan(r) / rho)
        u_d = p.fx * x * rad_dist_fac + p.cx
        v_d = p.fy * y * rad_dist_fac + p.cy
        return u_d, v_d, (theta < _find_theta_thresh_kannala(d)) & camera_in_front

    if d.model is DistortionModel.FISHEYE:
        rho = np.hypot(x, y)
        theta = np.arctan(rho)
        m = np.where(rho < epsilon, 1, theta / rho)
        t2 = theta * theta
        rad_dist_fac = np.polyval([d.k4, d.k3, d.k2, d.k1, 1], t2)
        u_d = p.fx * x * rad_dist_fac + p.cx
        v_d = p.fy * y * rad_dist_fac + p.cy
        return u_d, v_d, (theta < np.sqrt(_find_theta2_thresh(d))) & camera_in_front

    if (
        d.model is DistortionModel.BROWN_CONRADY
        or d.model is DistortionModel.OMNIDIRECTIONAL
    ):
        rsq = x * x + y * y
        x += d.lx
        y += d.ly
        rad_dist_fac = np.polyval([d.k3, d.k2, d.k1, 1], rsq)
        u_d = p.cx + p.fx * (
            x * rad_dist_fac + d.p1 * (rsq + 2 * x * x) + 2 * d.p2 * x * y
        )
        v_d = p.cy + p.fy * (
            y * rad_dist_fac + d.p2 * (rsq + 2 * y * y) + 3 * d.p1 * x * y
        )
        return u_d, v_d, (rsq < _find_rsq_thresh(d)) & camera_in_front

    if (
        d.model is DistortionModel.FISHEYE_RAD_TAN_THIN_PRISM
        or d.model is DistortionModel.FTHETA
        or d.model is DistortionModel.FISHEYE_RADIAL_CUSTOM
    ):
        raise Exception(f"Distortion model {d.model} is not supported")

    assert_never(d)



"""Pose class definition."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Collection, Optional, Type, Union

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation

from .common import Point3D, QuaternionData


@dataclass
class Pose:
    """A data class representing a pose in a 3D space. The pose includes both position and
    orientation.

    Attributes
    ----------
    heading : QuaternionData
        Orientation in the 3D space.
    position : Point3D
        Position in 3D space.
    """

    heading: QuaternionData
    position: Point3D

    def copy(self) -> Pose:
        """Creates a copy of this object.

        Returns
        -------
        Pose
            Clone of this object.
        """
        return Pose(heading=self.heading.copy(), position=self.position.copy())

    def as_euler(self, seq: str = "XYZ", degrees: bool = False) -> npt.NDArray:
        """Gets this object's heading as Euler angles.

        Parameters
        ----------
        seq : str, optional
            3 characters belonging to the set {'X', 'Y', 'Z'} for intrinsic rotations,
            or {'x', 'y', 'z'} for extrinsic rotations [1]. Adjacent axes cannot be the same.
            Extrinsic and intrinsic rotations cannot be mixed in one function call.
            By default "XYZ"
        degrees : bool, optional
            Whether angles are in degrees, by default False

        Returns
        -------
        npt.NDArray
            Vector of Euler angles.
        """
        return self.heading.as_euler(seq=seq, degrees=degrees)

    @staticmethod
    def from_euler(
        angles: npt.ArrayLike,
        seq: str = "XYZ",
        degrees: bool = False,
    ) -> Pose:
        """Create a transform from Euler angles.

        Parameters
        ----------
        angles : ArrayLike[float]
            Euler angles
        seq : str, optional
            3 characters belonging to the set {'X', 'Y', 'Z'} for intrinsic rotations,
            or {'x', 'y', 'z'} for extrinsic rotations [1]. Adjacent axes cannot be the same.
            Extrinsic and intrinsic rotations cannot be mixed in one function call.
            By default "XYZ"
        degrees : bool, optional
            Whether angles are in degrees, by default False

        Returns
        -------
        QuaternionData:
            Rotation as a Quaternion.
        """
        return Pose(
            heading=QuaternionData.from_euler(angles, seq=seq, degrees=degrees),
            position=Point3D.identity(),
        )

    @staticmethod
    def identity() -> Pose:
        """Represents no transformation

        Returns
        -------
        Pose:
            Identity rotation and translation pose.
        """
        return Pose(heading=QuaternionData.identity(), position=Point3D.identity())

    @staticmethod
    def from_rt(
        rotation: Optional[Union[QuaternionData, Rotation]] = None,
        translation: Optional[Union[Collection[float], Point3D]] = None,
    ) -> Pose:
        """Creates a Pose from a Rotation and a translation vector.

        Parameters
        ----------
        rotation : Union[Quaternion, QuaternionData, Rotation], optional
            A scipy Rotation object, by default None
        translation : Collection[float], optional
            Translation vector (x, y, z), by default None

        Returns
        -------
        Pose
            Object with rotation and translation aspects.
        """
        if rotation is None:
            rotation = QuaternionData.identity()
        elif isinstance(rotation, Rotation):
            rotation = QuaternionData.from_rotation(rotation)
        else:
            rotation = rotation.copy()

        if translation is None:
            translation = Point3D.identity()
        elif isinstance(translation, Point3D):
            translation = translation.copy()
        else:
            translation = Point3D(*translation)

        return Pose(
            position=translation,
            heading=rotation,
        )

    def as_transform(self) -> npt.NDArray[np.float64]:
        """Creates a rigid transformation matrix from a Pose.

        Returns
        -------
        npt.NDArray[np.float64]:
            A 4x4 matrix representing the rigid transformation matrix.
            ```py
            [
                [ r00, r01, r02, t0],
                [ r10, r11, r12, t1],
                [ r20, r21, r22, t2],
                [ 0, 0, 0, 1]
            ]
            ```
        """
        return np.block(
            [
                [
                    self.heading.as_matrix(),
                    np.asarray(self.position).reshape((3, 1)),
                ],
                [np.zeros(3), 1],
            ]
        )

    @staticmethod
    def from_transform(transform_matrix: npt.ArrayLike) -> Pose:
        """
        Creates a Pose from a rigid transformation matrix.

        Parameters
        ----------
        transform_matrix : npt.ArrayLike
            An object representing the rigid transformation matrix. Also accepts a 3x3 rotation
            matrix or a length 3 translation vector.
            ```py
            [
                [ r00, r01, r02, t0],
                [ r10, r11, r12, t1],
                [ r20, r21, r22, t2],
                [ 0, 0, 0, 1]
            ]
            ```

        Returns
        -------
        Pose
            A Pose object corresponding to the provided rigid transformation matrix.
        """
        transform_matrix = np.asarray(transform_matrix, dtype=np.float64)
        num_elements = np.prod(transform_matrix.shape)
        if num_elements == 16:
            transform_matrix = transform_matrix.reshape((4, 4))
            return Pose(
                heading=QuaternionData.from_matrix(transform_matrix[:3, :3]),
                position=Point3D(*transform_matrix[:3, 3]),
            )
        elif num_elements == 9:
            return Pose(
                heading=QuaternionData.from_matrix(transform_matrix),
                position=Point3D.identity(),
            )
        elif num_elements == 3:
            return Pose(
                heading=QuaternionData.identity(),
                position=Point3D(*transform_matrix.flatten()),
            )
        else:
            raise ValueError(f"Invalid shape {transform_matrix.shape}")

    @staticmethod
    def from_transformed_points(a: npt.ArrayLike, b: npt.ArrayLike) -> Pose:
        """Create a transform roughly mapping two point clouds together.
        Point clouds must have the same number of points
        and the point a[i] must correspond to point b[i].

        Parameters
        ----------
        a : npt.ArrayLike
            [[x, y, z], ...] Point cloud A
        b : npt.ArrayLike
            [[x, y, z], ...] Point cloud B

        Returns
        -------
        Pose
            The pose such that transform_points(pose, a) -> b
        """
        a = np.asarray(a, dtype=np.float64).reshape((-1, 3))
        b = np.asarray(b, dtype=np.float64).reshape((-1, 3))
        assert (
            tuple(a.shape) == tuple(b.shape)
        ), f"Point clouds do not have the same shape a={a.shape} b={b.shape}"

        centroid_a = np.mean(a, axis=0)
        centroid_b = np.mean(b, axis=0)
        relative_a = a - centroid_a
        relative_b = b - centroid_b

        c = relative_a.T @ relative_b
        v, s, w = np.linalg.svd(c)

        if (np.linalg.det(v) * np.linalg.det(w)) < 0.0:
            s[-1] = -s[-1]
            v[:, -1] = -v[:, -1]

        rotation = v @ w
        translation = centroid_b - centroid_a @ rotation

        return Pose.from_rt(Rotation.from_matrix(rotation.T), translation)

    @property
    def inverse(self) -> Pose:
        """Inverse of this object. When transforming points, the inverse will be the transformation
        in the opposite direction.

        Returns
        -------
        Pose
            Inverse of this object.
        """
        inverse_rotation = self.heading.as_rotation().inv()
        return Pose.from_rt(
            inverse_rotation, np.dot(-inverse_rotation.as_matrix(), self.position)
        )

    def __array__(
        self, dtype: Optional[Union[np.dtype, str, Type]] = None
    ) -> npt.NDArray:
        return np.array(
            [
                self.position.x,
                self.position.y,
                self.position.z,
                self.heading.x,
                self.heading.y,
                self.heading.z,
                self.heading.w,
            ],
            dtype=dtype,
        )

    def __eq__(self, other: Any) -> bool:
        if (
            hasattr(other, "heading")
            and hasattr(other, "position")
            and self.heading == other.heading
            and self.position == other.position
        ):
            return True
        us = self.__array__()
        other = np.asarray(other)
        return np.prod(us.shape) == np.prod(other.shape) and np.allclose(
            us, other.reshape(us.shape)
        )

    def __matmul__(self, other: Any) -> Any:
        if isinstance(other, Pose):
            return Pose.from_transform(self.as_transform() @ other.as_transform())
        return self.as_transform() @ other

    def __imatmul__(self, other: Any) -> Any:
        if isinstance(other, Pose):
            other = other.as_transform()
        result = Pose.from_transform(self.as_transform() @ other)
        self.heading = result.heading
        self.position = result.position
        return self

    def __repr__(self) -> str:
        return f"heading={self.heading} position={self.position}"



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, List, Literal, Optional, Sequence, TypeVar, Union

import scale_sensor_fusion_io as sfio
from typing_extensions import TypeAlias

PathField: TypeAlias = Union[int, str]
PathInput: TypeAlias = Union[PathField, List[PathField]]

T = TypeVar("T")


@dataclass
class ParseSuccess(Generic[T]):
    data: T
    success: Literal[True] = True


@dataclass
class ErrorDetails:
    path: List[PathField]
    errors: List[str]

    def __repr__(self) -> str:
        return f"{{ path: '{'.'.join([str(p) for p in self.path])}', errors: {self.errors} }}"

    @staticmethod
    def from_msg(msg: str, path: Optional[PathInput] = None) -> ErrorDetails:
        """
        Helper function to initiate a ErrorDetails from a single error message to reduce boilerplate
        """
        return ErrorDetails(
            path=(path if type(path) is list else [path]) if path else [], errors=[msg]  # type: ignore
        )

    @staticmethod
    def missing_field(field: str, path: Optional[PathInput] = None) -> ErrorDetails:
        """
        Helper function to template out details for missing field
        """
        return ErrorDetails(
            path=(path if type(path) is list else [path]) if path else [],  # type: ignore
            errors=[f"Missing field: {field}"],
        )


@dataclass
class DataValidationError(Exception):
    details: List[ErrorDetails]

    @staticmethod
    def from_msg(msg: str, path: Optional[PathInput] = None) -> DataValidationError:
        """
        Helper function to initiate a DataValidationError from a single error message to reduce boilerplate
        """
        return DataValidationError(details=[ErrorDetails.from_msg(msg, path)])

    def prepend_path(self, path: List[PathField]) -> DataValidationError:
        """Prepend path of error with additional prefix"""
        for err in self.details:
            err.path = path + err.path
        return self


@dataclass
class ParseError(DataValidationError):
    details: List[ErrorDetails]
    success: Literal[False] = False

    @staticmethod
    def from_msg(msg: str, path: Optional[PathInput] = None) -> ParseError:
        """
        Helper function to initiate a ParseError from a single error message to reduce boilerplate
        """
        return ParseError(details=[ErrorDetails.from_msg(msg, path)])

    @staticmethod
    def missing_field(field: str, path: Optional[PathInput] = None) -> ParseError:
        """
        Helper function to template out details for missing field
        """
        return ParseError(details=[ErrorDetails.missing_field(field, path)])


ParseResult: TypeAlias = Union[ParseSuccess[T], ParseError]

ValidationResult: TypeAlias = Optional[DataValidationError]



from dataclasses import dataclass
from typing import List, Literal, Optional

from ..common import AnnotationID, AnnotationKind, SensorID
from ..paths import AttributePath, CuboidPath


@dataclass
class CuboidActivationPath:
    sensor_id: SensorID
    timestamps: List[int]
    durations: List[float]
    cuboids: Optional[List[List[float]]]  # x y z yaw pitch roll dx dy dz


@dataclass
class CuboidProjectionPath:
    sensor_id: SensorID  # 2d sensor like cameras
    timestamps: List[int]
    boxes: List[
        Optional[List[float]]
    ]  # x y width height; Can be None if cuboid can't be projected or if projection is actively deleted
    cuboids: Optional[List[List[float]]]  # dx dy dz px py pz roll pitch yaw


@dataclass
class CuboidAnnotation:
    id: AnnotationID
    path: CuboidPath
    label: Optional[str] = None
    stationary: Optional[bool] = False
    type: Literal[AnnotationKind.Cuboid] = AnnotationKind.Cuboid
    attributes: Optional[List[AttributePath]] = None
    activations: Optional[List[CuboidActivationPath]] = None
    projections: Optional[List[CuboidProjectionPath]] = None
    parent_id: Optional[AnnotationID] = None



from dataclasses import dataclass, field
from typing import Dict, Any, Callable, Optional, Type, List, Union, Mapping


@dataclass
class Config:
    type_hooks: Mapping[Union[Type, object], Callable[[Any], Any]] = field(
        default_factory=dict
    )
    cast: List[Type] = field(default_factory=list)
    forward_references: Optional[Dict[str, Any]] = None
    check_types: bool = True
    strict: bool = False
    strict_unions_match: bool = False
    allow_missing_fields_as_none: bool = False



import pprint
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Callable, Generic, List, Literal, Optional, TypeVar, Union

import numpy as np
import scale_sensor_fusion_io.validation.dacite_internal as _dacite
from scale_json_binary import read_file
from scale_sensor_fusion_io.models import (
    CameraDistortion,
    CameraIntrinsics,
    CameraSensor,
    LidarSensor,
    LidarSensorFrame,
    LidarSensorPoints,
    PosePath,
    RadarSensor,
    RadarSensorFrame,
    RadarSensorPoints,
    Scene,
    Sensor,
)
from scale_sensor_fusion_io.spec import SFS
from scale_sensor_fusion_io.validation.error import (
    ErrorDetails,
    ParseError,
    ParseResult,
    ParseSuccess,
    PathField,
    ValidationResult,
)
from scale_sensor_fusion_io.validation.helpers import (
    convert_error,
    handle_dacite,
    is_strictly_increasing,
)
from typing_extensions import TypeAlias

CFG = _dacite.Config(cast=[Enum, tuple])

_T = TypeVar("_T")


"""Parse and validate a sfs file"""


def _handle_result(
    res: ParseResult[_T], error_details: List[ErrorDetails], path: List[PathField] = []
) -> Optional[_T]:
    if res.success:
        return res.data
    else:
        error_details.extend(
            res.details if not path else res.prepend_path(path).details
        )
        return None


def parse_radar(sensor: dict) -> ParseResult[SFS.RadarSensor]:
    error_details: List[ErrorDetails] = []

    parsed_sensor = handle_dacite(
        lambda: _dacite.from_dict(
            data_class=SFS.RadarSensor,
            data=sensor,
            config=CFG,
        ),
        error_details,
    )

    if parsed_sensor:
        if len(parsed_sensor.frames) <= 0:
            error_details.append(
                ErrorDetails(path=["frames"], errors=["Must have at least one frame"])
            )

        for idx, frame in enumerate(parsed_sensor.frames):
            """
            When reading via dacite, the numpy arrays aren't correctly shaped (since that's not included in the typedef)
            Thus, we reshape all fields here
            """
            frame.points.positions = frame.points.positions.reshape((-1, 3))
            frame.points.directions = (
                frame.points.directions.reshape((-1, 3))
                if frame.points.directions is not None
                else None
            )
            if (
                frame.points.timestamps is not None
                and frame.points.timestamps.dtype == np.uint64
            ):
                error_details.append(
                    ErrorDetails(
                        path=["frames", idx, "points", "timestamps"],
                        errors=["Uint64 timestamps are not supported yet"],
                    )
                )

    if len(error_details) > 0:
        return ParseError(details=error_details)

    assert parsed_sensor is not None
    return ParseSuccess(data=parsed_sensor)


def parse_lidar(sensor: dict) -> ParseResult[SFS.LidarSensor]:
    error_details: List[ErrorDetails] = []

    parsed_sensor = handle_dacite(
        lambda: _dacite.from_dict(
            data_class=SFS.LidarSensor,
            data=sensor,
            config=CFG,
        ),
        error_details,
    )

    if parsed_sensor:
        if len(parsed_sensor.frames) <= 0:
            error_details.append(
                ErrorDetails(path=["frames"], errors=["Must have at least one frame"])
            )

        for idx, frame in enumerate(parsed_sensor.frames):
            """
            When reading via dacite, the numpy arrays aren't correctly shaped (since that's not included in the typedef)
            Thus, we reshape all fields here
            """
            frame.points.positions = frame.points.positions.reshape((-1, 3))
            frame.points.colors = (
                frame.points.colors.reshape((-1, 3))
                if frame.points.colors is not None
                else None
            )
            if (
                frame.points.timestamps is not None
                and frame.points.timestamps.dtype == np.uint64
            ):
                error_details.append(
                    ErrorDetails(
                        path=["frames", idx, "points", "timestamps"],
                        errors=["Uint64 timestamps are not supported yet"],
                    )
                )

    if len(error_details) > 0:
        return ParseError(details=error_details)

    assert parsed_sensor is not None
    return ParseSuccess(data=parsed_sensor)


def parse_sensor(sensor: dict) -> ParseResult[SFS.Sensor]:
    error_details: List[ErrorDetails] = []

    # type
    sensor_type = sensor.get("type")
    if not sensor_type:
        error_details.append(ErrorDetails.missing_field("type"))

    parsed_sensor: Optional[SFS.Sensor] = None
    if sensor_type == "camera":
        parsed_sensor = handle_dacite(
            lambda: _dacite.from_dict(
                data_class=SFS.CameraSensor,
                data=sensor,
                config=CFG,
            ),
            error_details,
        )
    elif sensor_type == "lidar":
        parsed_sensor = _handle_result(parse_lidar(sensor), error_details)
    elif sensor_type == "radar":
        parsed_sensor = _handle_result(parse_radar(sensor), error_details)
    else:
        error_details.append(
            ErrorDetails(
                path=["type"], errors=[f"Invalid sensor type provided: {sensor_type}"]
            )
        )

    if len(error_details) > 0:
        return ParseError(details=error_details)

    assert parsed_sensor is not None
    return ParseSuccess(data=parsed_sensor)


def parse_scene_as_sfs(raw_data: dict) -> ParseResult[SFS.Scene]:
    """
    Parse raw dict as SFS.Scene

    Few notes:
      * We use a modified version of dacite to allow for aggregating errors instead of failing fast
      * We also don't run _dacite.from_dict on the scene object directly since it can't handle union types very elegantly currently
    """
    error_details: List[ErrorDetails] = []

    # version
    version = raw_data.get("version")
    if version is None:
        return ParseError.missing_field("version")

    if not raw_data["version"].startswith("1.0") and not raw_data["version"].startswith(
        "5.1"
    ):
        return ParseError.from_msg(
            f"Invalid version provided: {raw_data['version']}", path=["version"]
        )

    # sensors
    sensors = []
    _sensors = raw_data.get("sensors")
    if _sensors:
        if type(_sensors) != list:
            return ParseError.from_msg("Sensors must be a list", path=["sensors"])

        for idx, sensor in enumerate(_sensors):
            sensor = _handle_result(
                parse_sensor(sensor), error_details, path=["sensors", idx]
            )
            if sensor:
                sensors.append(sensor)

    # time_offset
    fields = SFS.Scene.__dataclass_fields__
    time_offset = handle_dacite(
        lambda: _dacite.cast_field(raw_data, field=fields["time_offset"], config=CFG),  # type: ignore
        error_details,
    )

    # time_unit
    time_unit = handle_dacite(
        lambda: _dacite.cast_field(raw_data, field=fields["time_unit"], config=CFG),  # type: ignore
        error_details,
    )

    # Additional scene level validations
    if len(error_details) == 0:
        scene = SFS.Scene(sensors=sensors, time_offset=time_offset, time_unit=time_unit)

        return ParseSuccess(data=scene)

    return ParseError(details=error_details)
    pass



"""Additional functionality for models.common"""
from typing import Iterable

import numpy as np
import numpy.typing as npt

from scale_sensor_fusion_io import Point3D


def point_3ds_to_matrix(points: Iterable[Point3D]) -> npt.NDArray[np.float64]:
    """Converts an iterable of Point3D objects into a matrix

    Parameters
    ----------
    points : Iterable[Point3D]
        Points to convert

    Returns
    -------
    npt.NDArray[np.float64]
        Matrix in the form [[x, y, z], ...] with the shape (len(points), 3)
    """
    return np.stack([np.asarray(point) for point in points])



import os
import numpy.typing as npt
import numpy as np
from typing import Any, List, Iterable, Optional, Union
from dataclasses import dataclass
from subprocess import Popen, PIPE
from tqdm import tqdm

import ffmpeg
from turbojpeg import TurboJPEG
import tempfile

turbo_jpeg = TurboJPEG()


DATA_DIR: str = (
    os.environ["DATA_DIR"]
    if "DATA_DIR" in os.environ
    else f'{os.environ.get("HOME", ".")}/data'
)


@dataclass
class VideoReader:
    video: Union[str, bytes]
    cache_dir: str
    threads: int = 0

    def __init__(
        self,
        video: Union[str, bytes],
        cache_dir: Optional[str] = None,
        threads: int = 0,
    ) -> None:
        self.video = video

        if cache_dir is None:
            if isinstance(video, str):
                with open(video, "rb") as f:
                    video_bytes = f.read()
            else:
                video_bytes = video

            self.cache_dir = os.path.join(
                tempfile.gettempdir(), f"video_reader_{hash(video_bytes)}"
            )
        else:
            self.cache_dir = cache_dir

        os.makedirs(self.cache_dir, exist_ok=True)

        self.threads = threads

    def build_cmd(self, video_file: str) -> List[str]:
        """Creates a command to extract frames from a video file path into
        an output directory using ffmpeg.

        After running the command, images will be outputted as
        `<self.cache_dir>/<1_index_frame_num>.jpg`.

        Parameters
        ----------
        video_file : str
            Path to the input video file

        Returns
        -------
        List[str]
            The ffpmeg command.
        """

        params = [
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel error",
            "-f mp4",
            f"-i {video_file}",
            os.path.join(self.cache_dir, "%d.jpg"),
        ]

        return [item for param in params for item in param.split(" ", 1)]

    def __get_or_write_video_file(self) -> str:
        """Gets self.video if it is a string, otherwise writes self.video to
        `<self.cache_dir>/video.mp4`.

        Parameters
        ----------
        cache_dir : str
            Directory to write the video file to, if self.video is not a string.

        Returns
        -------
        str
            The path to self.video.
        """

        if isinstance(self.video, str):
            return self.video

        video_file = os.path.join(self.cache_dir, "video.mp4")

        if not os.path.exists(video_file):
            with open(video_file, "wb") as fp:
                fp.write(self.video)

        return video_file

    def load(self, frame_num: int) -> npt.NDArray[np.uint8]:
        """Loads a frame from `self.video`.
        Saves video and frames with `self.cache_dir` if defined.
        Will fetch from `self.cache_dir` without loading if possible.

        Parameters
        ----------
        frame_num : int
            0 indexed frame number to return.

        Returns
        -------
        npt.NDArray[np.uint8]
            An image.

        Raises
        ------
        ValueError
            If ffmpeg fails to open.
        """

        image_file = os.path.join(self.cache_dir, f"{frame_num+1}.jpg")

        # Note: I tried to pipe through stdin, but was running into all kinda of errors
        # getting ffmpeg to dmux the raw bytes
        if not os.path.exists(image_file):
            video_file = self.__get_or_write_video_file()

            cmd = self.build_cmd(video_file)

            # start ffmpeg process
            with Popen(cmd, stdin=PIPE) as process:
                if process.stdin is None:
                    raise ValueError("Failed to open ffmpeg process")
                # wait for process to finish
                process.wait()

        with open(image_file, "rb") as fd:
            return turbo_jpeg.decode(fd.read(), 0)


@dataclass
class VideoWriter:
    target_file: str
    fps: Optional[int] = 10
    threads: Optional[int] = 0
    crf: Optional[int] = 24
    show_progress: Optional[bool] = True
    count: int = 0

    def build_cmd(self) -> List[str]:
        params = [
            "ffmpeg",
            "-y",
            "-hide_banner",
            "-loglevel error",
            "-f image2pipe",
            f"-r {self.fps}",
            "-i -",
            "-vcodec libx264",
            "-x264-params keyint=2:scenecut=0",
            "-pix_fmt yuv420p",
            f"-crf {self.crf}",
            f"-r {self.fps}",
            f"-threads {self.threads}",
            "-preset fast",
            self.target_file,
        ]

        return [item for param in params for item in param.split(" ", 1)]

    def encode(self, images: Iterable[Union[str, bytes]]) -> None:
        cmd = self.build_cmd()

        # start ffmpeg process
        process = Popen(cmd, stdin=PIPE)
        if process.stdin is None:
            raise ValueError("Failed to open ffmpeg process")

        if self.show_progress:
            images = tqdm(images)

        # stream images to ffmpeg
        for image in images:
            if isinstance(image, str):
                with open(image, "rb") as fp:
                    image = fp.read()

            process.stdin.write(image)
        process.stdin.close()

        # wait for process to finish
        process.wait()

    def get_video(self) -> bytes:
        with open(self.target_file, "rb") as fp:
            return fp.read()

    def writeFrame(self, im: npt.NDArray) -> None:
        self.jpeg_bytes.append(turbo_jpeg.encode(im, quality=95))
        self.count += 1

    def __enter__(self) -> "VideoWriter":
        self.jpeg_bytes: List[bytes] = []
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        if self.count > 0:
            self.encode(self.jpeg_bytes)


# Util function to generate a video from a list of images
def generate_video(
    image_files: List[str],
    target_file: str,
    fps: Optional[int] = 10,
    threads: Optional[int] = 0,
) -> None:
    encoder = VideoWriter(target_file, fps, threads)
    encoder.encode(image_files)


def write_audio_and_video(audio_file: str, video_file: str, output_file: str) -> None:
    if not os.path.isfile(audio_file) or not os.path.isfile(video_file):
        raise ValueError("Audio or video file does not exist")

    input_video = ffmpeg.input(video_file)
    input_audio = ffmpeg.input(audio_file)
    ffmpeg.concat(input_video, input_audio, v=1, a=1).output(
        output_file, loglevel="error"
    ).overwrite_output().run()



"""Contains ITransformablePath"""
from __future__ import annotations
from typing_extensions import Protocol

import numpy as np
import numpy.typing as npt
import pandas as pd


class ITransformablePath(Protocol):
    """A path that can be transformed using Pose or rigid transformation matrices."""

    @property
    def index(self) -> pd.Index:
        ...

    def as_matrix(self) -> npt.NDArray[np.float64]:
        ...

    @staticmethod
    def from_matrix(matrix: npt.ArrayLike, index: npt.ArrayLike) -> ITransformablePath:
        ...



from scale_sensor_fusion_io.model_converters.sfs import from_scene_spec_sfs
from .error import *
from .parser import *
from .validate import *


def parse_and_validate_scene(raw_scene: dict) -> ParseResult[Scene]:
    """
    1. parse dict to BS5.Scene
    2. convert to sfio.models.Scene
    3. run validations
    """
    parse_result = parse_scene_as_sfs(raw_scene)

    if not parse_result.success:
        return parse_result

    sfs_scene = parse_result.data

    try:
        scene = from_scene_spec_sfs(sfs_scene)
    except Exception as e:
        return ParseError.from_msg(
            f"Error converting to sfio.models.Scene. Error: {e}",
        )

    validate_result = validate_scene(scene)

    if not validate_result:
        # if no validate error, return parse result
        return ParseSuccess(data=scene)

    return ParseError(
        details=validate_result.details,
    )



import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation


# Utils
def points_to_df(points):
  positions = points['positions'].reshape(-1, 3)
  return pd.DataFrame({
    'x': positions[:, 0],
    'y': positions[:, 1],
    'z': positions[:, 2],
    'intensity': points['intensities'],
    'timestamp': points['timestamps'],
  })


def df_to_points(df):
  return {
    'positions': np.float32(df[['x', 'y', 'z']]),
    'intensities': np.uint8(df['intensity']),
    'timestamps': np.uint32(df['timestamp'])
  }


# Downsample functions
def voxel_downsample(df, voxel_size):
  df = df.sample(frac=1)
  df[['vx', 'vy', 'vz']] = (df[['x', 'y', 'z']] / voxel_size).astype(int)
  return df.groupby(['vx', 'vy', 'vz']).mean()


def voxel_rand_downsample(df, voxel_size):
  df = df.sample(frac=1)
  df[['vx', 'vy', 'vz']] = (df[['x', 'y', 'z']] / voxel_size).astype(int)
  return df.groupby(['vx', 'vy', 'vz']).first()


def hypervoxel_rand_downsample(df, voxel_size, time_window):
  df = df.sample(frac=1)
  df[['vx', 'vy', 'vz']] = (df[['x', 'y', 'z']] / voxel_size).astype(int)
  df['vt'] = (df['timestamp'] / time_window).astype(int)
  return df.groupby(['vx', 'vy', 'vz', 'vt']).first()


def random_downsample(df, ratio: float):
  return df.sample(int(len(df) * ratio))


# Scene downsample
def take_timestamp(elem):
    return elem['timestamp']


def downsample_scene(scene: dict, downsampler):
  total_points = 0
  downsampled_points = 0
  for sensor in scene['sensors']:
    if sensor['type'] == 'lidar':
      sensor['frames'] = sorted(sensor['frames'][:200], key=take_timestamp)
      for frame in sensor['frames']:
        points_df = points_to_df(frame['points'])
        downsampled_df = downsampler(points_df)
        frame['points'] = df_to_points(downsampled_df)
        total_points += len(points_df)
        downsampled_points += len(downsampled_df)
  print(f"Downsampled {total_points} to {downsampled_points} ({downsampled_points/total_points:.04f})")
  return scene



"""Additional functionality for using PosePath."""

from typing import Optional, Union

import numpy as np
import numpy.typing as npt
import pandas as pd

from scale_sensor_fusion_io import FillValue, PosePath
from scale_sensor_fusion_io.utils.pose_helpers import transform_points


def apply_interpolated_transform_to_points(
    points: npt.ArrayLike,
    timestamps: npt.ArrayLike,
    pose_path_adjustment: PosePath,
    resolution: Optional[Union[float, int]] = None,
    extrapolation_strategy: Optional[str] = None,
) -> npt.NDArray:
    """
    Applies interpolated transformations from the given PosePath to the given points based on their
    corresponding timestamps and the specified resolution.

    This method groups points that have timestamps closer than the provided resolution value and
    applies the same transformation to each point within the group, improving performance by
    reducing the number of separate interpolations and transformations required.

    Parameters
    ----------
    points : npt.ArrayLike
        An array-like object containing the 3D points to be transformed, with shape (N, 3),
        where N is the number of points.
    timestamps : npt.ArrayLike
        An array-like object containing the timestamps corresponding to each point in the points
        array, with shape (N,).
    pose_path_adjustment : PosePath
        This PosePath is used to transform the points. Transforms are interpolated from supplied
        timestamps.
    resolution : float, optional, default: None
        The time resolution for grouping points. Points will be bucketed by this resolution and be
        adjusted based off of the average timestamp of the group. If not provided, will not group
        points and will interpolate each point using its timestamp
    extrapolation_strategy : str, optional, default: None
        Only supports "linear" extrapolation.

    Returns
    -------
    npt.NDArray
        A numpy array containing the transformed points, with shape (N, 3).
    """
    fill_value: FillValue = (
        "extrapolate" if extrapolation_strategy == "linear" else "nearest"
    )

    if resolution is None:
        return per_point_motion_compensation(
            points, timestamps, pose_path_adjustment, fill_value
        )
    else:
        return bucketed_motion_compensation(
            points, timestamps, pose_path_adjustment, fill_value, resolution
        )


def bucketed_motion_compensation(
    points: npt.ArrayLike,
    timestamps: npt.ArrayLike,
    pose_path_adjustment: PosePath,
    fill_value: FillValue = "nearest",
    resolution: Union[float, int] = 1000,
) -> npt.NDArray:
    """Computes motion compensation for points by bucketing them by timestamp
    and applying the same transformation to each point within the bucket.
    """
    points = np.asarray(points, dtype=np.float64).reshape((-1, 3))
    timestamps = np.asarray(timestamps).reshape((-1))

    # Create intervals and groups
    pd_timestamps = pd.Series(timestamps)
    intervals = (pd_timestamps / resolution).astype(int)
    groups = pd_timestamps.groupby(intervals)

    # Get groups timestamps and interpolate poses
    interval_timestamps = groups.mean()

    # Resample poses to generate the transforms
    transforms = pose_path_adjustment.interpolate(
        interval_timestamps, fill_value=fill_value
    ).as_matrix()

    # Prepare an empty array to store the transformed points
    # This is done to preserve order of points
    transformed_points = np.empty_like(points)

    for index, [_, group] in enumerate(groups):
        transformed_group = transform_points(transforms[index], points[group.index])
        transformed_points[group.index] = transformed_group

    return transformed_points


def per_point_motion_compensation(
    points: npt.ArrayLike,
    timestamps: npt.ArrayLike,
    pose_path_adjustment: PosePath,
    fill_value: FillValue = "nearest",
) -> npt.NDArray:
    """Applies per-point motion compensation to points."""
    transform_path = pose_path_adjustment.interpolate(timestamps, fill_value=fill_value)
    return apply_pose_path_to_points(transform_path, points)


def apply_pose_path_to_points(
    pose_path: PosePath, points: npt.ArrayLike
) -> npt.NDArray:
    """Transforms points using the poses of a pose path, matching pose_path[i] with points[i].

    Parameters
    ----------
    pose_path : PosePath
        The poses to use as transforms. Must be the same length as points.
    points : npt.ArrayLike
        An array-like object of points with shape (N, 3), where N must be the same as the number
        of poses.

    Returns
    -------
    npt.NDArray
        An array of transformed points with shape (N, 3).
    """
    points = np.asarray(points, dtype=np.float64).reshape((-1, 3))
    assert len(pose_path) == len(points), (
        "PosePath and points do not have the same length. "
        f"len(pose_path)={len(pose_path)} len(points)={len(points)}"
    )
    return np.array(
        [
            transform_points(t, [points[i]])[0]
            for i, t in enumerate(pose_path.as_matrix())
        ]
    )


def apply_pose_path_to_single_point(
    pose_path: PosePath, point: npt.ArrayLike
) -> npt.NDArray:
    """Transforms a single point using the poses of a pose path.

    Parameters
    ----------
    pose_path : PosePath
        The poses to use as transforms.
    point : npt.ArrayLike
        An array-like object of points with shape (N, 3), where N is the number of points.

    Returns
    -------
    npt.NDArray
        An array of transformed points with shape (N, 3), where N is the number of poses in
        pose_path.
    """
    points = np.asarray(point, dtype=np.float64).reshape((1, 3))
    return np.array([transform_points(t, points)[0] for t in pose_path.as_matrix()])



"""
This file has type checking/casting functions used within dacite that are not publicly surfaced
but are useful to use for our parsing function
"""

from dataclasses import InitVar
from typing import Type, Any, Optional, Union, Collection, TypeVar, Dict, Callable, Mapping, List, Tuple, get_type_hints

T = TypeVar("T", bound=Any)


def transform_value(
    type_hooks: Mapping[Union[Type, object], Callable[[Any], Any]], cast: List[Type], target_type: Type, value: Any
) -> Any:
    # Generic hook type match
    if Any in type_hooks:
        value = type_hooks[Any](value)
    if is_generic_collection(target_type):
        collection_type = extract_origin_type(target_type)
        if collection_type and collection_type in type_hooks:
            value = type_hooks[collection_type](value)
    # Exact hook type match
    if target_type in type_hooks:
        value = type_hooks[target_type](value)
    else:
        # Cast to types in cast list
        for cast_type in cast:
            if is_subclass(target_type, cast_type):
                if is_generic_collection(target_type):
                    origin_collection = extract_origin_collection(target_type)
                    if is_set(origin_collection):
                        return list(value)
                    value = origin_collection(value)
                else:
                    value = target_type(value)
                break
    # Peel optional types
    if is_optional(target_type):
        if value is None:
            return None
        target_type = extract_optional(target_type)
        return transform_value(type_hooks, cast, target_type, value)
    # For collections (dict/list), transform each item
    if is_generic_collection(target_type) and isinstance(value, extract_origin_collection(target_type)):
        collection_cls = value.__class__
        if issubclass(collection_cls, dict):
            key_cls, item_cls = extract_generic(target_type, defaults=(Any, Any))
            return collection_cls(
                {
                    transform_value(type_hooks, cast, key_cls, key): transform_value(type_hooks, cast, item_cls, item)
                    for key, item in value.items()
                }
            )
        item_cls = extract_generic(target_type, defaults=(Any,))[0]
        return collection_cls(transform_value(type_hooks, cast, item_cls, item) for item in value)
    return value


def get_data_class_hints(data_class: Type[T], globalns: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    type_hints = get_type_hints(data_class, globalns=globalns)
    for attr, type_hint in type_hints.items():
        if is_init_var(type_hint):
            type_hints[attr] = extract_init_var(type_hint)
    return type_hints


def extract_origin_collection(collection: Type) -> Type:
    try:
        return collection.__extra__
    except AttributeError:
        return collection.__origin__


def extract_origin_type(collection: Type) -> Optional[Type]:
    collection_type = extract_origin_collection(collection)
    if collection_type is list:
        return List
    elif collection_type is dict:
        return Dict
    return None


def is_optional(type_: Type) -> bool:
    return is_union(type_) and type(None) in extract_generic(type_)


def extract_optional(optional: Type[Optional[T]]) -> T:
    other_members = [member for member in extract_generic(optional) if member is not type(None)]
    if other_members:
        return Union[tuple(other_members)]  # type: ignore
    else:
        raise ValueError("can not find not-none value")


def is_generic(type_: Type) -> bool:
    return hasattr(type_, "__origin__")


def is_union(type_: Type) -> bool:
    if is_generic(type_) and type_.__origin__ == Union:
        return True

    try:
        from types import UnionType  # type: ignore

        return isinstance(type_, UnionType)
    except ImportError:
        return False


def is_tuple(type_: Type) -> bool:
    return is_subclass(type_, tuple)


def is_literal(type_: Type) -> bool:
    try:
        from typing import Literal  # type: ignore

        return is_generic(type_) and type_.__origin__ == Literal
    except ImportError:
        return False


def is_new_type(type_: Type) -> bool:
    return hasattr(type_, "__supertype__")


def extract_new_type(type_: Type) -> Type:
    return type_.__supertype__


def is_init_var(type_: Type) -> bool:
    return isinstance(type_, InitVar) or type_ is InitVar


def is_set(type_: Type) -> bool:
    return type_ in (set, frozenset) or isinstance(type_, (frozenset, set))


def extract_init_var(type_: Type) -> Union[Type, Any]:
    try:
        return type_.type
    except AttributeError:
        return Any


def is_instance(value: Any, type_: Type) -> bool:
    if type_ == Any:
        return True
    elif is_union(type_):
        return any(is_instance(value, t) for t in extract_generic(type_))
    elif is_generic_collection(type_):
        origin = extract_origin_collection(type_)
        if not isinstance(value, origin):
            return False
        if extract_generic_no_defaults(type_) is None:
            return True
        if isinstance(value, tuple) and is_tuple(type_):
            tuple_types = extract_generic(type_)
            if len(tuple_types) == 1 and tuple_types[0] == ():
                return len(value) == 0
            elif len(tuple_types) == 2 and tuple_types[1] is ...:
                return all(is_instance(item, tuple_types[0]) for item in value)
            else:
                if len(tuple_types) != len(value):
                    return False
                return all(is_instance(item, item_type) for item, item_type in zip(value, tuple_types))
        if isinstance(value, Mapping):
            key_type, val_type = extract_generic(type_, defaults=(Any, Any))
            for key, val in value.items():
                if not is_instance(key, key_type) or not is_instance(val, val_type):
                    return False
            return True
        return all(is_instance(item, extract_generic(type_, defaults=(Any,))[0]) for item in value)
    elif is_new_type(type_):
        return is_instance(value, extract_new_type(type_))
    elif is_literal(type_):
        return value in extract_generic(type_)
    elif is_init_var(type_):
        return is_instance(value, extract_init_var(type_))
    elif is_type_generic(type_):
        return is_subclass(value, extract_generic(type_)[0])
    elif is_generic(type_):
        origin = extract_origin_collection(type_)
        return isinstance(value, origin)
    else:
        try:
            # As described in PEP 484 - section: "The numeric tower"
            if isinstance(value, (int, float)) and type_ in [float, complex]:
                return True
            return isinstance(value, type_)
        except TypeError:
            return False


def is_generic_collection(type_: Type) -> bool:
    if not is_generic(type_):
        return False
    origin = extract_origin_collection(type_)
    try:
        return bool(origin and issubclass(origin, Collection) and not skip_generic_conversion(origin))
    except (TypeError, AttributeError):
        return False


def skip_generic_conversion(origin: Type) -> bool:
    return origin.__module__ == "numpy" and origin.__qualname__ == "ndarray"


def extract_generic(type_: Type, defaults: Tuple = ()) -> tuple:
    try:
        if hasattr(type_, "_special") and type_._special:
            return defaults
        return type_.__args__ or defaults
    except AttributeError:
        return defaults


def extract_generic_no_defaults(type_: Type) -> Union[tuple, None]:
    try:
        if hasattr(type_, "_special") and type_._special:
            return None
        return type_.__args__
    except AttributeError:
        return None


def is_subclass(sub_type: Type, base_type: Type) -> bool:
    if is_generic_collection(sub_type):
        sub_type = extract_origin_collection(sub_type)
    try:
        return issubclass(sub_type, base_type)
    except TypeError:
        return False


def is_type_generic(type_: Type) -> bool:
    try:
        return type_.__origin__ in (type, Type)
    except AttributeError:
        return False



from dataclasses import dataclass
from typing import List, Literal, Optional, Union, Sequence

import numpy as np
import numpy.typing as npt

from ...common import SensorID, SensorKind
from ...paths import PosePath


# Define PointsSensorPoints dataclass
@dataclass
class Points:
    positions: npt.NDArray[np.float32]
    colors: Optional[npt.NDArray[np.uint8]] = None


# Define PointsSensor dataclass
@dataclass
class PointsSensor:
    id: SensorID
    points: Points
    parent_id: Optional[SensorID] = None
    type: Literal[SensorKind.Points] = SensorKind.Points



from enum import Enum
from typing import cast

import dacite
from scale_json_binary import JSONBinaryEncoder
import scale_sensor_fusion_io.models
from scale_sensor_fusion_io.model_converters import from_scene_spec_sfs

from ..spec import SFS

encoder = JSONBinaryEncoder()


def _fix_data_shape(scene: SFS.Scene) -> SFS.Scene:
    """
    When reading via dacite, the numpy arrays aren't correctly shaped (since that's not included in the typedef)
    This function fixes all the fields that need to be reshaped
    """
    if scene.sensors:
        for sensor in scene.sensors:
            if sensor.type == "lidar":
                for l_frame in sensor.frames:
                    l_frame.points.positions = l_frame.points.positions.reshape(-1, 3)
                    if l_frame.points.colors is not None:
                        l_frame.points.colors = l_frame.points.colors.reshape(-1, 3)

            elif sensor.type == "radar":
                for r_frame in sensor.frames:
                    r_frame.points.positions = r_frame.points.positions.reshape(-1, 3)
                    if r_frame.points.directions is not None:
                        r_frame.points.directions = r_frame.points.directions.reshape(
                            -1, 3
                        )
                    if r_frame.points.lengths is not None:
                        r_frame.points.lengths = r_frame.points.lengths.reshape(-1, 3)

            elif sensor.type == "points":
                sensor.points.positions = sensor.points.positions.reshape(-1, 3)
    return scene


class SFSLoader:
    def __init__(
        self,
        scene_url: str,
    ):
        self.scene_url = scene_url

    def load(self) -> scale_sensor_fusion_io.models.Scene:
        scene_sfs = self.load_as_sfs()
        return from_scene_spec_sfs(scene_sfs)
    
    @staticmethod
    def load_from_bytes(raw_data: bytes) -> SFS.Scene:
        obj = encoder.loads(raw_data)
        if "version" not in obj or not obj["version"].startswith("1.0"):
            raise Exception(f"Cannot load scene with version {obj['version']}")

        scene_bs5 = dacite.from_dict(
            data_class=SFS.Scene,
            data=obj,
            config=dacite.Config(
                cast=[Enum, tuple],
            ),
        )

        scene = _fix_data_shape(scene_bs5)

        return scene

    def load_as_sfs(self) -> SFS.Scene:
        raw_data: bytes
        with open(self.scene_url, "rb") as fd:
            raw_data = cast(bytes, fd.read())

        return self.load_from_bytes(raw_data)
       

    def load_unsafe(self) -> dict:
        """
        Loads the scene as a typed dict without doing any validation or parsing. It just hackily casts to the TypedDict representation of the scene spec.

        This is primarily useful for doing quick scripting where you may want to fix previously created, malformed scene
        """
        pass

        with open(self.scene_url, "rb") as fd:
            raw_data = cast(bytes, fd.read())

        obj = encoder.loads(raw_data)
        if "version" not in obj or not obj["version"].startswith("1.0"):
            raise Exception(f"Cannot load scene with version {obj['version']}")

        return obj



from dataclasses import dataclass
from typing import Any, Dict, List, Literal, Optional, Union

import numpy as np
import numpy.typing as npt

SensorID = Union[str, int]
AnnotationID = Union[str, int]


# Define PosePath dataclass
@dataclass
class PosePath:
    timestamps: List[int]
    values: List[List[float]]  # x y z qx qy qz qw


# Define PointsSensorPoints dataclass
@dataclass
class PointsSensorPoints:
    positions: npt.NDArray[np.float32]
    colors: Optional[npt.NDArray[np.uint8]]


# Define PointsSensor dataclass
@dataclass
class PointsSensor:
    id: SensorID
    points: PointsSensorPoints
    type: Literal["points"] = "points"
    parent_id: Optional[SensorID] = None


# Define LidarSensorPoints dataclass
@dataclass
class LidarSensorPoints:
    positions: npt.NDArray[np.float32]
    colors: Optional[npt.NDArray[np.uint8]] = None
    intensities: Optional[npt.NDArray[np.uint8]] = None
    timestamps: Optional[Union[npt.NDArray[np.uint32], npt.NDArray[np.uint64]]] = None


# Define LidarSensorFrame dataclass
@dataclass
class LidarSensorFrame:
    timestamp: int
    points: LidarSensorPoints


# Define LidarSensor dataclass
@dataclass
class LidarSensor:
    id: SensorID
    poses: PosePath
    frames: List[LidarSensorFrame]
    parent_id: Optional[SensorID] = None
    coordinates: Literal["ego", "world"] = "world"
    type: Literal["lidar"] = "lidar"


# Define RadarSensorPoints dataclass
@dataclass
class RadarSensorPoints:
    positions: npt.NDArray[np.float32]
    directions: Optional[npt.NDArray[np.float32]] = None
    lengths: Optional[npt.NDArray[np.float32]] = None
    timestamps: Optional[Union[npt.NDArray[np.uint32], npt.NDArray[np.uint64]]] = None


# Define RadarSensorFrame dataclass
@dataclass
class RadarSensorFrame:
    timestamp: int
    points: RadarSensorPoints


# Define RadarSensor dataclass
@dataclass
class RadarSensor:
    id: SensorID
    poses: PosePath
    frames: List[RadarSensorFrame]
    type: Literal["radar"] = "radar"
    coordinates: Literal["ego", "world"] = "world"
    parent_id: Optional[SensorID] = None


# Define DistortionModel enum
DistortionModel = Literal[
    "brown_conrady",
    "mod_equi_fish",
    "mod_kannala",
    "fisheye",
    "fisheye_rad_tan_prism",
    "cylindrical",
    "equirectangular",
    "omnidirectional",
    "fisheye_radial_custom",
    "ftheta",
]


# Define CameraDistortion dataclass
@dataclass
class CameraDistortion:
    model: DistortionModel
    params: List[float]


# Define CameraIntrinsics dataclass
@dataclass
class CameraIntrinsics:
    fx: float
    fy: float
    cx: float
    cy: float
    width: int
    height: int
    distortion: Optional[CameraDistortion]


# Define CameraSensorContent dataclass
@dataclass
class CameraSensorVideo:
    timestamps: List[int]
    content: npt.NDArray[np.uint8]
    fps: float


# Define CameraSensorImages dataclass
@dataclass
class CameraSensorImage:
    timestamp: int
    content: npt.NDArray[np.uint8]


# Define CameraSensor dataclass
@dataclass
class CameraSensor:
    id: SensorID
    poses: PosePath
    intrinsics: CameraIntrinsics
    video: Optional[CameraSensorVideo] = None
    images: Optional[List[CameraSensorImage]] = None
    type: Literal["camera"] = "camera"
    parent_id: Optional[SensorID] = None


# Define OdometrySensor dataclass
@dataclass
class OdometrySensor:
    id: SensorID
    poses: PosePath
    type: Literal["odometry"] = "odometry"
    parent_id: Optional[SensorID] = None


# Define AttributePath dataclass
@dataclass
class AttributePath:
    name: str
    timestamps: List[int]
    values: List[Union[str, int, List[str]]]
    sensor_id: Optional[SensorID] = None
    static: bool = False


# Define CuboidPath dataclass
@dataclass
class CuboidPath:
    timestamps: List[int]
    values: List[List[float]]  # dx dy dz px py pz roll pitch yaw


# Define CuboidActivationPath dataclass
@dataclass
class CuboidActivationPath:
    sensor_id: SensorID
    timestamps: List[int]
    durations: List[float]
    cuboids: Optional[List[List[float]]]  # dx dy dz px py pz roll pitch yaw


# Define CuboidProjectionPath dataclass
@dataclass
class CuboidProjectionPath:
    sensor_id: SensorID  # 2d sensor like cameras
    timestamps: List[int]
    boxes: List[
        Optional[List[float]]
    ]  # x y width height; Can be None if cuboid can't be projected or if projection is actively deleted
    cuboids: Optional[List[List[float]]]  # dx dy dz px py pz roll pitch yaw


# Define CuboidAnnotation dataclass
@dataclass
class CuboidAnnotation:
    id: AnnotationID
    path: CuboidPath
    label: Optional[str] = None
    stationary: Optional[bool] = False
    type: Literal["cuboid"] = "cuboid"
    attributes: Optional[List[AttributePath]] = None
    activations: Optional[List[CuboidActivationPath]] = None
    projections: Optional[List[CuboidProjectionPath]] = None
    parent_id: Optional[AnnotationID] = None


# Define AttributesAnnotation dataclass
@dataclass
class AttributesAnnotation:
    id: AnnotationID
    type: Literal["attributes"] = "attributes"
    parent_id: Optional[AnnotationID] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class AnnotationPath:
    timestamps: List[int]
    values: List[List[float]]


@dataclass
class Point2DAnnotation:
    id: AnnotationID
    sensor_id: SensorID
    path: AnnotationPath  # values: [x, y]
    type: Literal["point_2d"] = "point_2d"
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class Box2DAnnotation:
    id: AnnotationID
    sensor_id: SensorID
    path: AnnotationPath  # values: [left, top, width, height]
    type: Literal["box_2d"] = "box_2d"
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


class Polyline2DAnnotation:
    id: AnnotationID
    sensor_id: SensorID
    path: AnnotationPath  # x_0, y_0, x_1, y_1, ..., x_n, y_n
    type: Literal["polyline_2d"] = "polyline_2d"
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class Polygon2DAnnotation:
    id: AnnotationID
    sensor_id: SensorID
    path: AnnotationPath  # x_0, y_0, x_1, y_1, ..., x_n, y_n
    type: Literal["polygon_2d"] = "polygon_2d"
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class PolygonAnnotation:
    id: AnnotationID
    sensor_id: Optional[SensorID]
    path: AnnotationPath  # x_0, y_0, z_0, x_1, y_1, z_1, ..., x_n, y_n, z_n
    type: Literal["polygon"] = "polygon"
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class TopdownPolygonAnnotation:
    id: AnnotationID
    sensor_id: Optional[SensorID]
    path: AnnotationPath  # x_0, y_0, z_0, x_1, y_1, z_1, ..., x_n, y_n, z_n
    type: Literal["polygon_topdown"] = "polygon_topdown"
    parent_id: Optional[AnnotationID] = None
    stationary: Literal[True] = True
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class PolylineAnnotation:
    id: AnnotationID
    sensor_id: Optional[SensorID]
    path: AnnotationPath  # x_0, y_0, z_0, x_1, y_1, z_1, ..., x_n, y_n, z_n
    type: Literal["polyline"] = "polyline"
    is_closed: Optional[bool] = None
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None

@dataclass
class TopdownPointAnnotation:
    id: AnnotationID
    sensor_id: Optional[SensorID]
    path: AnnotationPath  # x_0, y_0, z_0, x_1, y_1, z_1, ..., x_n, y_n, z_n
    type: Literal["point_topdown"] = "point_topdown"
    parent_id: Optional[AnnotationID] = None
    stationary: Literal[True] = True
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class PointAnnotation:
    id: AnnotationID
    sensor_id: Optional[SensorID]
    path: AnnotationPath  # x, y, z
    type: Literal["point"] = "point"
    parent_id: Optional[AnnotationID] = None
    stationary: Optional[bool] = False
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class EventAnnotation:
    id: AnnotationID
    start: int
    type: Literal["event"] = "event"
    parent_id: Optional[AnnotationID] = None
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None
    duration: Optional[int] = None
    sensor_id: Optional[SensorID] = None


# Define LabeledPointsAnnotationLabeledPoint dataclass
@dataclass
class LabeledPoint:
    sensor_id: SensorID
    point_ids: npt.NDArray[np.uint32]
    sensor_frame: Optional[int] = None


# Define LabeledPointsAnnotation dataclass
@dataclass
class LabeledPointsAnnotation:
    id: AnnotationID
    label: str
    labeled_points: List[LabeledPoint]
    is_instance: bool = False
    type: Union[Literal["labeled_points"], Literal["point_segmentation"]] = "point_segmentation"
    parent_id: Optional[AnnotationID] = None


@dataclass
class LocalizationAdjustmentAnnotation:
    id: AnnotationID
    poses: PosePath
    type: Literal["localization_adjustment"] = "localization_adjustment"
    parent_id: Optional[AnnotationID] = None


@dataclass
class LinkAnnotation:
    id: AnnotationID
    sensor_id: Optional[SensorID]
    label: str
    is_bidirectional: bool
    from_id: AnnotationID
    to_id: AnnotationID
    type: Literal["link"] = "link"
    parent_id: Optional[AnnotationID] = None
    attributes: Optional[List[AttributePath]] = None


@dataclass
class ObjectAnnotation:
    id: AnnotationID
    type: Literal["object"] = "object"
    parent_id: Optional[AnnotationID] = None
    label: Optional[str] = None
    attributes: Optional[List[AttributePath]] = None


Sensor = Union[CameraSensor, LidarSensor, RadarSensor, OdometrySensor, PointsSensor]

Annotation = Union[
    CuboidAnnotation,
    AttributesAnnotation,
    Box2DAnnotation,
    Point2DAnnotation,
    Polyline2DAnnotation,
    Polygon2DAnnotation,
    PolygonAnnotation,
    TopdownPolygonAnnotation,
    TopdownPointAnnotation,
    Polyline2DAnnotation,
    PolylineAnnotation,
    PointAnnotation,
    LinkAnnotation,
    LabeledPointsAnnotation,
    LocalizationAdjustmentAnnotation,
    ObjectAnnotation,
    EventAnnotation,
]

# Define Scene dataclass


@dataclass
class Scene:
    version: Literal["1.0"] = "1.0"
    sensors: Optional[List[Sensor]] = None
    annotations: Optional[List[Annotation]] = None
    attributes: Optional[List[AttributePath]] = None
    time_offset: Optional[int] = None
    time_unit: Optional[Literal["microseconds", "nanoseconds"]] = "microseconds"
    metadata: Optional[Dict[str, Any]] = None



"""Contains InterpolateablePath ABC"""
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Type, TypeVar

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy.interpolate import interp1d
from scipy.spatial.transform import Rotation, Slerp


from ..common import FillValue


@dataclass
class SpacialInterpolationData:
    """Values to use for interpolation

    Members
    -------
    linear : Optional[npt.NDArray], optional, default: None
    spherical : Optional[Rotation], optional, default: None
    """

    linear: Optional[npt.NDArray] = None
    spherical: Optional[Rotation] = None

    def __repr__(self) -> str:
        return f"(linear={self.linear} spherical={self.spherical})"


class SpacialPath(ABC, pd.DataFrame):
    """A path containing 3D spacial data."""

    @staticmethod
    def __lerp(
        x: npt.NDArray,
        y: Optional[npt.NDArray],
        t: npt.NDArray,
        x_sorted_order: Optional[npt.NDArray[np.int32]],
        t_sorted_order: Optional[npt.NDArray[np.int32]],
        t_unsorted_order: Optional[npt.NDArray[np.int32]],
        fill_value: FillValue,
    ) -> Optional[npt.NDArray]:
        """Helper function for SpacialPath.interpolate. Handles linear interpolation."""
        if y is None:
            return None

        if x_sorted_order is not None:
            y = y[x_sorted_order]

        if t_sorted_order is not None:
            t = t[t_sorted_order]

        if fill_value == "nearest":
            interpolator_fill_value = (y[0], y[-1])
        elif fill_value == "identity":
            interpolator_fill_value = np.zeros(6)
        else:
            interpolator_fill_value = fill_value

        lerp = interp1d(
            x=x, y=y, bounds_error=False, fill_value=interpolator_fill_value, axis=0  # type: ignore
        )
        result = lerp(t)

        if t_unsorted_order is not None:
            result = result[t_unsorted_order]
        return result

    @staticmethod
    def __slerp(
        x: npt.NDArray,
        y: Optional[Rotation],
        t: npt.NDArray,
        x_sorted_order: Optional[npt.NDArray[np.int32]],
        t_sorted_order: Optional[npt.NDArray[np.int32]],
        t_unsorted_order: Optional[npt.NDArray[np.int32]],
        fill_value: FillValue,
    ) -> Optional[Rotation]:
        """Helper function for SpacialPath.interpolate. Handles spherical linear interpolation."""
        if y is None or len(x) == 0:
            return None

        if x_sorted_order is not None:
            y = y[x_sorted_order]

        if t_sorted_order is not None:
            t = t[t_sorted_order]

        lower_bound = np.searchsorted(t, x[0], side="left")
        upper_bound = np.searchsorted(t, x[-1], side="right")

        slerp = Slerp(x, y)
        rotations = []
        if lower_bound > 0:
            angles = SpacialPath.__lerp(
                x, y.as_euler("xyz"), t[:lower_bound], None, None, None, fill_value
            )
            if angles is not None:
                rotations.append(
                    Rotation.from_euler(
                        "xyz",
                        angles,
                    )
                )

        if lower_bound < upper_bound:
            rotations.append(slerp(t[lower_bound:upper_bound]))

        if upper_bound < len(t):
            angles = SpacialPath.__lerp(
                x, y.as_euler("xyz"), t[upper_bound:], None, None, None, fill_value
            )
            if angles is not None:
                rotations.append(
                    Rotation.from_euler(
                        "xyz",
                        angles,
                    )
                )

        result = Rotation.concatenate(rotations)

        if t_unsorted_order is not None and result is not None:
            result = result[t_unsorted_order]
        return result

    def interpolate(
        self: SpacialPathT,
        index: npt.ArrayLike,
        fill_value: FillValue = "nearest",
    ) -> SpacialPathT:
        """Interpolate a 3D path object at the given indices.

        Parameters
        ----------
        path : SpacialPathT
            3D spacial data (rotation + translation) over time.
        index : npt.ArrayLike
            Timestamps to interpolate the path at.
        fill_value : FillValue, optional, default: "nearest"
            How to handle out of bound data.

        Returns
        -------
        SpacialPathT
            A path object with values interpolated at the given indices.
        """
        index = np.asarray(index).flatten()
        if np.array_equal(self.index, index):
            return self.copy()

        if len(self.index) == 0:
            raise ValueError("Cannot interpolate with empty path")
        if len(self.index) == 1:
            return type(self)(self.take([0] * len(index)).values, index=index)

        # index and path.index need to be sorted
        if np.any(index[1:] < index[:-1]):
            input_sorted_order = np.argsort(index)
            input_unsorted_order = np.argsort(input_sorted_order)
        else:
            input_sorted_order = None
            input_unsorted_order = None

        x = self.index.to_numpy().flatten()
        y = self.get_interpolation_data()
        if np.any(x[1:] < x[:-1]):
            x_sorted_order = np.argsort(x)
            x = x[x_sorted_order]
        else:
            x_sorted_order = None

        values = SpacialInterpolationData(
            linear=SpacialPath.__lerp(
                x,
                y.linear,
                index,
                x_sorted_order,
                input_sorted_order,
                input_unsorted_order,
                fill_value,
            ),
            spherical=SpacialPath.__slerp(
                x,
                y.spherical,
                index,
                x_sorted_order,
                input_sorted_order,
                input_unsorted_order,
                fill_value,
            ),
        )

        return type(self).from_interpolation_data(values, index=index)

    @classmethod
    @abstractmethod
    def from_interpolation_data(
        cls: Type[SpacialPathT],
        value: SpacialInterpolationData,
        index: npt.NDArray,
    ) -> SpacialPathT:
        ...

    @abstractmethod
    def get_interpolation_data(self) -> SpacialInterpolationData:
        ...


SpacialPathT = TypeVar("SpacialPathT", bound=SpacialPath)
